''**************************************************************************
'' Definitions useful for FAT12/FAT16 implementation
''**************************************************************************

#ifndef FS_FAT_D_INC
#define FS_FAT_D_INC

'' FAT types (these values are indexes in the array of size TOTAL_FATS)
#define FAT12		0
#define FAT16		1
#define FAT32		2
#define TOTAL_FATS	3

#define	FAT12_NAME	"FAT12"
#define	FAT16_NAME	"FAT16"

#define	FAT12_NAME_EXT	"FAT12   "
#define	FAT16_NAME_EXT	"FAT16   "

'---------------------------------------------------------------------------
'		MS-DOS FAT -- File Allocation Table Structures
'---------------------------------------------------------------------------

#define MB_PARTITION_TYPE_FAT12		01h
#define MB_PARTITION_TYPE_FAT16		06h

#define MB_PARTITION_BOOTABLE		80h
#define MB_PARTITION_NOT_BOOTABLE	00h

#define MB_PART_1_INFO_OFFS 		1beh
#define MB_PART_2_INFO_OFFS 		1ceh
#define MB_PART_3_INFO_OFFS 		1deh
#define MB_PART_4_INFO_OFFS 		1eeh

'' #define BOOT_MAGIC		055AAh		' "U"
#define BOOT_MAGIC		0AA55h		' "U"
#define BOOT_MAGIC_POS		01FEh		'
#define BOOT_MAGIC_SIZE		2

#define MAX_CLUSTERS_FAT12	4087
#define	BP_NUM_OF_FAT		2
#define	BP_MEDIA_ID_ANY		0f8h
#define	BP_JUMP_INSTR		"eb 3e 90"%
#define	BP_OEM_NAME		"BTIGER  "

'---------------------------------------------------------------------------
' Various Definitions
'---------------------------------------------------------------------------
#define FS_INFO_MAGIC	     	61417272h	' "rrAa"
#define FS_INFO_MAGIC_POS    	01E4h		'

#define EXT_FLAGS_FAT_MASK   	000Fh		'
#define EXT_FLAGS_FAT_MIRROR 	0080h		'

#define FAT_EMPTY_CLUSTER	00h

#define FAT12_MIN_CLUSTERs   	00000001h	' 1
#define FAT12_MAX_CLUSTERs   	00000FF4h	' 00000FFFh - 11	' 4084
#define FAT12_BAD	     	00000FF7h	'
#define FAT12_EOF	     	00000FF8h	'
#define FAT12_EOF_MAX	     	00000FFFh	'
#define FAT12_MULT		3		' to avoid operations with reals
#define FAT12_DIV		2		' we use x*3/2 instead of x*1.5

#define FAT16_MIN_CLUSTERs   	00000FF5h	' 00000FFFh - 10	' 4085
#define FAT16_MAX_CLUSTERs   	0000FFF4h	' 0000FFFFh - 11	' 65524
#define FAT16_BAD	     	0000FFF7h	'
#define FAT16_EOF	     	0000FFF8h	'
#define FAT16_EOF_MAX	     	0000FFFFh	'
#define FAT16_MULT		2		' to avoid operations with reals in FAT12
#define FAT16_DIV		1		'

#define FAT32_MIN_CLUSTERs   	0000FFF5h	' 0000FFFFh - 10	' 65525
#define FAT32_MAX_CLUSTERs   	0FFFFFF4h	' 0FFFFFFFh - 11	' 268435444
#define FAT32_BAD	     	0FFFFFF7h	'
#define FAT32_EOF	     	0FFFFFF8h	'
#define FAT32_EOF_MAX	     	0FFFFFFFh	'
#define FAT32_MULT		4		' to avoid operations with reals in FAT12
#define FAT32_DIV		1		'

#define DIR_NOT_USED_ID		00h
#define DIR_ERASED_ID		0E5h
#define DIR_ERASED_ID_STR	"E5"%
#define DIR_E5_MASK_ID		05h
#define DIR_DOT_ID		2Eh

#define DIR_ATTR_FILE		000000b
#define DIR_ATTR_READONLY	000001b
#define DIR_ATTR_SYSTEM		000010b
#define DIR_ATTR_HIDDEN		000100b
#define DIR_ATTR_VOLUME		001000b
#define DIR_ATTR_DIRECTORY	010000b
#define DIR_ATTR_ARCHIVE	100000b

#define	DIR_NAME_SIZE		11
#define	ALTERNATE_DIR_NAME_SIZE	14
#define	DIR_EXT_POS		8
#define	DIR_EXT_SIZE		3
#define	DIR_EXT_TERM		"."
#define	DIR_NAME_FULL_SIZE	12
#define	DIR_NAME_FILL_PAT	" "
#define	DIR_NAME_CONT_CHAR	"~"

#define BOOT_RECORD_SIZE	512
#define DIR_ENTRY_SIZE		32
#define DIR_ENTRIES_NUM		256
#define DIR_ENTRIES_NUM_WIN	512
#define READ_FAT_BUFFER_SIZE	128

#ifdef DIR_ENTRIES_IN_RAM
'' DIR_ENTRY_IN_RAM_SIZE = DIR_ENTRY_SIZE + Address (Long) + Flags (Long)
#define DIR_ENTRY_IN_RAM_SIZE	40
'' it would be good to keep this value at approx. TOTAL_FILE_NUMBER * 2
#define DIR_ENTRIES_IN_RAM_NUM	32
'' additional positions in the ram buffer for dir entries
#define DIR_IN_RAM_ADDRESS_POS	20h
#define DIR_IN_RAM_FLAGS_POS	24h
#define DIR_IN_RAM_ADDRESS_SIZE	4
#define DIR_IN_RAM_FLAGS_SIZE	4
#endif

#define PATH_LIM_DOS		"\"
#define PATH_LIM_UNIX		"/"

'---------------------------------------------------------------------------
' MBR for FAT12	and for FAT16 volumes
'---------------------------------------------------------------------------
#define MBR_X86BOOT_POS			00h
#define MBR_START_HEAD_POS		01h
#define MBR_START_SECT_POS		02h
#define MBR_START_CYLINDER_POS		03h
#define MBR_PARTITION_TYPE_POS		04h
#define MBR_END_HEAD_POS		05h
#define MBR_END_SECT_POS		06h
#define MBR_END_CYLINDER_POS		07h
#define MBR_LOG_START_SECT_POS		08h
#define MBR_NUM_SECT_POS		0Ch

#define MBR_X86BOOT_SIZE		1
#define MBR_START_HEAD_SIZE		1
#define MBR_START_SECT_SIZE		1
#define MBR_START_CYLINDER_SIZE		1
#define MBR_PARTITION_TYPE_SIZE		1
#define MBR_END_HEAD_SIZE		1
#define MBR_END_SECT_SIZE		1
#define MBR_END_CYLINDER_SIZE		1
#define MBR_LOG_START_SECT_SIZE		4
#define MBR_NUM_SECT_SIZE		4

'---------------------------------------------------------------------------
' BPB for FAT12	and for FAT16 volumes
'---------------------------------------------------------------------------
#define BPB_BYTES_PER_SECT_POS		0Bh
#define BPB_SECT_PER_CLUSTER_POS	0Dh
#define BPB_RESERVED_SECT_POS		0Eh
#define BPB_NUMBER_OF_FATS_POS		10h
#define BPB_ROOT_ENTRIES_POS		11h
#define BPB_TOTAL_SECT_POS		13h
#define BPB_MEDIA_POS			15h
#define BPB_SECT_PER_FAT_POS		16h
#define BPB_SECT_PER_TRACK_POS		18h
#define BPB_HEADS_PER_CYL_POS		1Ah
#define BPB_HIDDEN_SECT_POS		1Ch
#define BPB_TOTAL_SECT_BIG_POS		20h

#define BPB_BYTES_PER_SECT_SIZE		2
#define BPB_SECT_PER_CLUSTER_SIZE	1
#define BPB_RESERVED_SECT_SIZE		2
#define BPB_NUMBER_OF_FATS_SIZE		1
#define BPB_ROOT_ENTRIES_SIZE		2
#define BPB_TOTAL_SECT_SIZE		2
#define BPB_MEDIA_SIZE			1
#define BPB_SECT_PER_FAT_SIZE		2
#define BPB_SECT_PER_TRACK_SIZE		2
#define BPB_HEADS_PER_CYL_SIZE		2
#define BPB_HIDDEN_SECT_SIZE		4
#define BPB_TOTAL_SECT_BIG_SIZE		4


'---------------------------------------------------------------------------
' BPB for FAT32 volumes (additional fields)
'---------------------------------------------------------------------------
#define BPB32_SECT_PER_FAT_POS		24h
#define BPB32_FLAGS_POS			28h
#define BPB32_VERSION_POS		2Ah
#define BPB32_ROOT_CLUSTER_POS		2Ch
#define BPB32_INFO_SECT_POS		30h
#define BPB32_BOOT_BACKUP_START_POS	32h
#define BPB32_RESERVED_POS		34h

#define BPB32_SECT_PER_FAT_SIZE		4
#define BPB32_FLAGS_SIZE		2
#define BPB32_VERSION_SIZE		2
#define BPB32_ROOT_CLUSTER_SIZE		4
#define BPB32_INFO_SECT_SIZE		2
#define BPB32_BOOT_BACKUP_START_SIZE	2
#define BPB32_RESERVED_SIZE		12


'---------------------------------------------------------------------------
' Boot Sector layout for FAT12, FAT16, FAT32
'---------------------------------------------------------------------------
#define BS_JMP_POS			00h
#define BS_OEM_NAME_POS			03h
#define BS_BPB_POS			0Bh
#define BS_DRIVE_NUMBER_POS		24h
#define BS_UNUSED_POS			25h
#define BS_EXT_BOOT_SIGN_POS		26h
#define BS_SERIAL_NUMBER_POS		27h
#define BS_VOLUME_LABEL_POS		2Bh
#define BS_FILE_SYSTEM_POS		36h
#define BS_BOOT_CODE_POS		3Eh

#define BS_JMP_SIZE			3
#define BS_OEM_NAME_SIZE		8
#define BS_BPB_SIZE			25
#define BS_DRIVE_NUMBER_SIZE		1
#define BS_UNUSED_SIZE			1
#define BS_EXT_BOOT_SIGN_SIZE		1
#define BS_SERIAL_NUMBER_SIZE		4
#define BS_VOLUME_LABEL_SIZE		11
#define BS_FILE_SYSTEM_SIZE		8
#define BS_BOOT_CODE_SIZE		450

#define BS32_DRIVE_NUMBER_POS		40h
#define BS32_UNUSED_POS			41h
#define BS32_EXT_BOOT_SIGN_POS		42h
#define BS32_SERIAL_NUMBER_POS		43h
#define BS32_VOLUME_LABEL_POS		47h
#define BS32_FILE_SYSTEM_POS		52h
#define BS32_BOOT_CODE_POS		5Ah

#define BS32_DRIVE_NUMBER_SIZE		1
#define BS32_UNUSED_SIZE		1
#define BS32_EXT_BOOT_SIGN_SIZE		1
#define BS32_SERIAL_NUMBER_SIZE		4
#define BS32_VOLUME_LABEL_SIZE		11
#define BS32_FILE_SYSTEM_SIZE		8
#define BS32_BOOT_CODE_SIZE		422


'---------------------------------------------------------------------------
' FAT32 File System Information Structure
'---------------------------------------------------------------------------
#define FS_INFO_SIGNATURE_POS		00h
#define FS_INFO_FREE_CLUSTER_COUNT_POS	04h
#define FS_INFO_NEXT_CLUSTER_POS	08h
#define FS_INFO_RESERVED_POS		0Ch


'---------------------------------------------------------------------------
' Directory Entry format
'---------------------------------------------------------------------------
#define DIR_SPEC_INFO_POS	00h
#define DIR_NAME_POS		00h
#define DIR_ATTR_POS		0Bh
#define DIR_RESERVED_POS	0Ch
#define DIR_CREATE10MS_POS	0Dh
#define DIR_CREATETIME_POS	0Eh
#define DIR_CREATEDATE_POS	10h
#define DIR_ACCESSDATE_POS	12h
#define DIR_CLUSTERHIGH_POS	14h
#define DIR_TIME_POS		16h
#define DIR_DATE_POS		18h
#define DIR_CLUSTER_POS		1Ah
#define DIR_SIZE_POS		1Ch


'---------------------------------------------------------------------------
' LFN (Long File Name) Entry format
'---------------------------------------------------------------------------
#define LFN_SEQUENCE_POS	00h
#define LFN_NAME1_POS		01h
#define LFN_ATTR_POS		0Bh
#define LFN_RESERVED_POS	0Ch
#define LFN_CHECKSUM_POS	0Dh
#define LFN_NAME2_POS		0Eh
#define LFN_CLUSTER_POS		1Ah
#define LFN_NAME3_POS		1Ch


'---------------------------------------------------------------------------
' LFN Specific Definitions
'---------------------------------------------------------------------------
#define LFN_MAX_NAME			251		' trial-and-failure (Win95)
#define LFN_MAX_PATH			260		'

#define LFN_TAG_ATTR			0Fh
#define LFN_TAG_MASK			0Fh

#define LFN_SEQ_DELETED			80h
#define LFN_SEQ_LAST			40h
#define LFN_SEQ_MASK			3Fh

#define LFN_LETTERS_PER_SLOT		13

'' Pattern byte to fill up the sectors
#define FAT_FILL_SECT_PAT_STR		"00"%

'' Pattern byte to fill up erased sectors
#define FAT_FILL_SECT_ERASE_PAT_STR	"ff"%

'' Max size of path
#define	MAX_PATH_SIZE			256

#define MAX_LONGNAME_SIZE		256

'' Max size of block for testing whether area is free or not
#define MAX_TEST_BLOCK_SIZE		256

#define	PATH_NAME_TYPE_SHORT		0
#define	PATH_NAME_TYPE_LONG		1
#define	PATH_NAME_TYPE_FOR_SEARCH	2

#define	FAT_INSTR_FUNC_BUF_SIZE		128

#endif

