''**************************************************************************
'' Implementation of FAT12/FAT16 with long names support
''**************************************************************************

#ifndef FS_FAT_I_INC
#define FS_FAT_I_INC
'' Master Boot for FAT12 and for FAT16 volumes
byte bMbx86boot		'' x86 defaut boot partition
byte bMbStartHead	'' start head
byte bMbStartSector	'' start sector
byte bMbStartCylinder	'' start cylinder
byte bMbPartitionType	'' partition type
byte bMbEndHead		'' end head
byte bMbEndSector	'' end sector
byte bMbEndCylinder	'' end cylinder
long lMbLogStartSector	'' logical start sector
long lMbNumSectors	'' number of all sectors
word wMbSignature	'' signature word


'' BPB for FAT12 and for FAT16 volumes
word wBpbBytesPerSector
byte bBpbSectorsPerCluster
word wBpbReservedSectors
byte bBpbNumberOfFATs
word wBpbRootEntries
long lBpbTotalSectors
byte bBpbMedia
word wBpbSectorsPerFAT
word wBpbSectorsPerTrack
word wBpbHeadsPerCylinder
long lBpbHiddenSectors
long lBpbTotalSectorsBig


'' BPB for FAT32 volumes (additional fields)
long lBpb32SectorsPerFAT
word wBpb32Flags
word wBpb32Version
long lBpb32RootCluster
word wBpb32InfoSector
word wBpb32BootBackupStart
string abBpb32Reserved$(12)


'' Boot Sector layout for FAT12, FAT16, FAT32
string sBsJmp$(BS_JMP_SIZE)
string sBsOemName$(BS_OEM_NAME_SIZE)
byte bBsDriveNumber
byte bBsUnused
byte bBsExtBootSignature
long lBsSerialNumber
string sBsVolumeLabel$(BS_VOLUME_LABEL_SIZE)
string sBsFileSystem$(BS_FILE_SYSTEM_SIZE)
string sBsBootCode$(BS_BOOT_CODE_SIZE)


'' FAT32 File System Information Structure
long lFsInfoSignature
long lFsInfoFreeClusterCount
long lFsInfoNextCluster
string sFsInfoReserved$(12)


'' Directory Entry format
string sDirName$(DIR_NAME_SIZE)
byte bDirAttr
byte bDirReserved
byte bDirCreate10MS
word wDirCreateTime
word wDirCreateDate
word wDirAccessDate
word wDirClusterHigh
word wDirTime
word wDirDate
word wDirCluster
long lDirSize


'' LFN (Long File Name) Entry format
byte bLfnSequence
string sLfnName1$(10)
byte bLfnAttr
byte bLfnReserved
byte bLfnChecksum
string sLfnName2$(12)
word wLfnCluster
string sLfnName3$(4)

long lFATSectors, lDataSectors, lRootSectors, lTotalClusters
long lBytesPerCluster, lBytesPerFAT
long lFATStart, lRootAreaStart
byte bFATType
'' Two variables to avoid operations with reals,
'' while calculating the position of a cluster inside of FAT
byte bFATClusterPosMult, bFATClusterPosDiv

'' Buffer to temporary save data of a cluster
string sClusterBuffer$(MAX_CLUSTER_SIZE)

array alFATEndClusterMinId(TOTAL_FATS) of long
array alFATEndClusterMaxId(TOTAL_FATS) of long
array alFATInvalidClusterId(TOTAL_FATS) of long

'' address of current dir in the FAT
string sCurrentDir$( MAX_PATH_SIZE )
long lCurrentDirFATAddr

'' if this flag set to TRUE, use windows-like formatting parameters
byte bFATFmtWinFlag

#ifdef FAT_IN_RAM
'' buffer to temporary keep the file allocation table
array asFATRamBuffer$(FAT_RAM_BUF_NUM) of string (FAT_RAM_BUF_SIZE)
string sFATTempBuffer$(FAT_RAM_BUF_SIZE)
byte bFATTableModified
array abFATTableModifiedAreas(FAT_MODIFIED_AREAS_MAX_NUM) of byte
long lFATModifiedAreasNum
long lFATModifiedAreaSize
#endif

#ifdef DIR_ENTRIES_IN_RAM
array asFATDirEntryRamBuffer$(DIR_ENTRIES_IN_RAM_NUM) of string (DIR_ENTRY_IN_RAM_SIZE)
byte bFATDirEntryRamBufModified
long lFATDirEntryRamBufSize
#endif


''**************************************************************************
'' Buffering of the file allocation table
''
sub lFATReadTableNum( long lpAddr; byte bpSize; var long lpvNum; var byte bpvIsOk )
#ifdef FAT_IN_RAM
  long llBufIndex, llInBufAddr
  llBufIndex = ( lpAddr - lFATStart ) / FAT_RAM_BUF_SIZE
  llInBufAddr = mod( lpAddr - lFATStart, FAT_RAM_BUF_SIZE )
  lpvNum = nfroms( asFATRamBuffer$(llBufIndex), llInBufAddr, bpSize )
  bpvIsOk = TRUE
#else
  call lHdReadNum( lpAddr, bpSize, lpvNum, bpvIsOk )
#endif
end

'' restriction: it is not possible that the string to read comes over
'' two or more limits of the temporary FAT buffer
sub sFATReadTableString( long lpAddr; long lpSize; var string spvBuffer$; var byte bpvIsOk )
#ifdef FAT_IN_RAM
  long llBufIndex, llInBufAddr, llSize
  llBufIndex = ( lpAddr - lFATStart ) / FAT_RAM_BUF_SIZE
  llInBufAddr = mod( lpAddr - lFATStart, FAT_RAM_BUF_SIZE )
  if llInBufAddr + lpSize <= FAT_RAM_BUF_SIZE then
    spvBuffer$ = mid$( asFATRamBuffer$(llBufIndex), llInBufAddr, lpSize )
  else
    llSize = FAT_RAM_BUF_SIZE - llInBufAddr
    spvBuffer$ = mid$( asFATRamBuffer$(llBufIndex), llInBufAddr, llSize )
    spvBuffer$ = stos$( spvBuffer$, llSize, asFATRamBuffer$(llBufIndex+1), 0, lpSize - llSize )
  endif
  bpvIsOk = TRUE
#else
  call sHdReadString( lpAddr, lpSize, spvBuffer$, bpvIsOk )
#endif
end


'' restriction: it is not possible that the string to write comes over
'' two or more limits of the temporary FAT buffer
sub vFATWriteTableString( long lpAddr, lpPos, lpSize; var string spvBuffer$ )
#ifdef FAT_IN_RAM
  long llBufIndex, llInBufAddr, llSize, llAreaIndex, llArea2Index
  llBufIndex = ( lpAddr - lFATStart ) / FAT_RAM_BUF_SIZE
  llInBufAddr = mod( lpAddr - lFATStart, FAT_RAM_BUF_SIZE )
  if llInBufAddr + lpSize <= FAT_RAM_BUF_SIZE then
    asFATRamBuffer$(llBufIndex) = stos$( asFATRamBuffer$(llBufIndex), llInBufAddr, &
      spvBuffer$, lpPos, lpSize )
  else
    llSize = FAT_RAM_BUF_SIZE - llInBufAddr
    asFATRamBuffer$(llBufIndex) = stos$( asFATRamBuffer$(llBufIndex), llInBufAddr, &
      spvBuffer$, lpPos, llSize )
    asFATRamBuffer$(llBufIndex+1) = stos$( asFATRamBuffer$(llBufIndex+1), 0, &
      spvBuffer$, llSize, lpSize - llSize )
  endif
  bFATTableModified = TRUE
  '' set modified area flag
  llAreaIndex = ( lpAddr - lFATStart ) / lFATModifiedAreaSize
  if llAreaIndex < lFATModifiedAreasNum then
    abFATTableModifiedAreas(llAreaIndex) = TRUE
  endif
  llArea2Index = ( lpAddr - lFATStart + lpSize ) / lFATModifiedAreaSize
  if llArea2Index <> llAreaIndex and llArea2Index < lFATModifiedAreasNum then
    abFATTableModifiedAreas(llArea2Index) = TRUE
  endif
#else
  long llAddrInFAT2
  call vHdWriteString( lpAddr, lpPos, lpSize, spvBuffer$ )
  '' update the copy of FAT also
  llAddrInFAT2 = lpAddr + lBytesPerFAT
  call vHdWriteString( llAddrInFAT2, lpPos, lpSize, spvBuffer$ )
#endif
end


#ifdef FAT_IN_RAM
sub bFATInitTable( var byte bpvIsOk )
  long llBufNum, llBufIndex, llCurBufSize, llSizeRead, llNextAddr, llIndex
  bpvIsOk = FALSE
  if lBytesPerFAT <= MAX_FAT_SIZE then
    llBufNum = lBytesPerFAT / FAT_RAM_BUF_SIZE
    for llBufIndex = 0 to FAT_RAM_BUF_NUM - 1
      asFATRamBuffer$(llBufIndex) = ""
    next
    llBufIndex = 0
    llSizeRead = 0
    while llSizeRead < lBytesPerFAT
      if lBytesPerFAT - llSizeRead > FAT_RAM_BUF_SIZE then
        llCurBufSize = FAT_RAM_BUF_SIZE
      else
        llCurBufSize = lBytesPerFAT - llSizeRead
      endif
      llNextAddr = lFATStart + llSizeRead
      call sHdReadString( llNextAddr, llCurBufSize, asFATRamBuffer$(llBufIndex), bpvIsOk )
      llSizeRead = llSizeRead + llCurBufSize
      llBufIndex = llBufIndex + 1
    endwhile
    if lBytesPerFAT / FAT_MODIFIED_AREAS_MAX_NUM >= lHdBytesInPage then
      lFATModifiedAreaSize = lBytesPerFAT / FAT_MODIFIED_AREAS_MAX_NUM
      lFATModifiedAreaSize = lFATModifiedAreaSize + lHdBytesInPage - &
        mod( lFATModifiedAreaSize, lHdBytesInPage )
    else
      lFATModifiedAreaSize = lHdBytesInPage
    endif
    lFATModifiedAreasNum = lBytesPerFAT / lFATModifiedAreaSize
    if mod( lBytesPerFAT, lFATModifiedAreaSize ) > 0 then
      lFATModifiedAreasNum = lFATModifiedAreasNum + 1
    endif
    for llIndex = 0 to lFATModifiedAreasNum - 1
      abFATTableModifiedAreas(llIndex) = FALSE
    next
    bFATTableModified = FALSE
  endif
end


sub vFATSynchronizeTable()
  long llNextFATAddr, llNextFAT2Addr, llBufNum, llBufIndex, llCurBufSize, &
    llSizeWritten, llAreaIndex, llPrevBufIndex

  if bFATTableModified = FALSE then
    return
  endif
  llPrevBufIndex = -1
  for llAreaIndex = 0 to lFATModifiedAreasNum
    if abFATTableModifiedAreas(llAreaIndex) = TRUE then
      '' calculate to which FAT ram buffer belongs the area
      llBufIndex = llAreaIndex * lFATModifiedAreaSize / FAT_RAM_BUF_SIZE
      if llBufIndex <> llPrevBufIndex then
    	llSizeWritten = llBufIndex * FAT_RAM_BUF_SIZE
        if lBytesPerFAT - llSizeWritten > FAT_RAM_BUF_SIZE then
          llCurBufSize = FAT_RAM_BUF_SIZE
    	else
          llCurBufSize = lBytesPerFAT - llSizeWritten
    	endif
    	llNextFATAddr = lFATStart + llSizeWritten
    	call vHdWriteString( llNextFATAddr, 0, llCurBufSize, asFATRamBuffer$(llBufIndex) )
    	llNextFAT2Addr = llNextFATAddr + lBytesPerFAT
    	call vHdWriteString( llNextFAT2Addr, 0, llCurBufSize, asFATRamBuffer$(llBufIndex) )
      endif
      llPrevBufIndex = llBufIndex
    endif
  next
  bFATTableModified = FALSE
end


'' The sub returns the id of the next cluster in cluster chain of FAT;
'' If next cluster cannot be found, the return value "lpvNextClusterAddr" is -1
sub lFATTableGetNextClusterAddr( long lpCurClusterAddr, lpRepeatNum; var long lpvNextClusterAddr )
  byte blIsValid, blIsEnd, blIsOk
  long llCurAddr, llIndx, llClusterAddr
  long llBufIndex, llInBufAddr
  string slBuffer$(READ_FAT_BUFFER_SIZE)

  call bFATIsEndCluster( lpCurClusterAddr, blIsEnd )
  if blIsEnd = TRUE then
    lpvNextClusterAddr = -1
    return
  endif

  lpvNextClusterAddr = lpCurClusterAddr
  llClusterAddr = lpCurClusterAddr

  llIndx = 0
  while llIndx < lpRepeatNum and lpvNextClusterAddr <> -1
    llCurAddr = llClusterAddr * bFATClusterPosMult / bFATClusterPosDiv

    llBufIndex = llCurAddr / FAT_RAM_BUF_SIZE
    llInBufAddr = mod( llCurAddr, FAT_RAM_BUF_SIZE )
    lpvNextClusterAddr = nfroms( asFATRamBuffer$(llBufIndex), llInBufAddr, 4 )

    switch bFATType
      case FAT12:
        lpvNextClusterAddr = lpvNextClusterAddr bitand 0ffffh
        if mod( llClusterAddr, 2 ) = 0 then
          lpvNextClusterAddr = lpvNextClusterAddr bitand 0fffh
        else
          lpvNextClusterAddr = ( lpvNextClusterAddr shr 4 ) bitand 0fffh
        endif
      case FAT16:
        lpvNextClusterAddr = lpvNextClusterAddr bitand 0ffffh
      case FAT32:
    endswitch

    call bFATIsClusterValid( llClusterAddr, blIsValid )
    if blIsValid = FALSE then
      lpvNextClusterAddr = -1
    endif
    llIndx = llIndx + 1
    llClusterAddr = lpvNextClusterAddr
  endwhile
end


sub lFATTableGetUsedClusters( var long lpvUsedClusters )
  long llBufIndex, llBadClusters, llI, llJ, llLen, llInstrLen
  string slInstrBuffer$( FAT_INSTR_FUNC_BUF_SIZE )

#ifdef FS_SECURE_INIT
  slInstrBuffer$ = "" '$JK
#endif	'' FS_SECURE_INIT

  lpvUsedClusters = 0
  llBadClusters = 0
  if bFATType = FAT12 then
  else
    if bFATType = FAT16 then
      for llBufIndex = 0 to FAT_RAM_BUF_NUM - 1
        if asFATRamBuffer$(llBufIndex) <> "" then
          '' remove all empty clusters
          sFATTempBuffer$ = remdouble$( asFATRamBuffer$(llBufIndex), "<00>" )
          llLen = len( sFATTempBuffer$ )
          if llLen > 0 then
            '' count all bad clusters ("FFF7")
            llI = 0
            llInstrLen = llLen
            if llInstrLen > FAT_INSTR_FUNC_BUF_SIZE then
              llInstrLen = FAT_INSTR_FUNC_BUF_SIZE
            endif
            while llI < llLen
              llJ = 0
              slInstrBuffer$ = mid$( sFATTempBuffer$, llI, llInstrLen )
              while llJ < llInstrLen
                llJ = instr( "<0FFh><0F7h>", slInstrBuffer$, 0, 4 )
                if llJ >= 0 then
                  llJ = llJ + 2
                  llBadClusters = llBadClusters + 1
                else
                  llJ = llInstrLen
                endif
              endwhile
              llI = llI + llInstrLen
              if llI + FAT_INSTR_FUNC_BUF_SIZE > llLen then
                llInstrLen = llLen - llI
              else
                llInstrLen = FAT_INSTR_FUNC_BUF_SIZE
              endif
            endwhile
            lpvUsedClusters = lpvUsedClusters + len( sFATTempBuffer$ ) / 2
          endif
        endif
      next
    endif
  endif
  lpvUsedClusters = lpvUsedClusters - llBadClusters
end
#endif
''
''**************************************************************************


''**************************************************************************
'' Buffering of the directory entries
''
sub sFATReadDirEntryString( long lpAddr; long lpSize; var string spvBuffer$; var byte bpvIsOk )
#ifdef DIR_ENTRIES_IN_RAM
  long llIndex, llNextFreeIndex
  call lFATGetDirEntryInRamIndex( lpAddr, llIndex, llNextFreeIndex )
  if llIndex < 0 then
    call sHdReadString( lpAddr, lpSize, spvBuffer$, bpvIsOk )
  else
    spvBuffer$ = mid$( asFATDirEntryRamBuffer$(llIndex), 0, DIR_ENTRY_SIZE )
  endif
#else
  call sHdReadString( lpAddr, lpSize, spvBuffer$, bpvIsOk )
#endif
end


sub vFATWriteDirEntryString( long lpAddr, lpPos, lpSize; var string spvBuffer$ )
#ifdef DIR_ENTRIES_IN_RAM
  long llIndex, llNextFreeIndex
  call lFATGetDirEntryInRamIndex( lpAddr, llIndex, llNextFreeIndex )
  if llIndex < 0 then
    if llNextFreeIndex < 0 then
      '' the ram buffer for dir entries is full, write the string down to the media
      call vHdWriteString( lpAddr, lpPos, lpSize, spvBuffer$ )
    else
      asFATDirEntryRamBuffer$(llNextFreeIndex) = stos$( asFATDirEntryRamBuffer$(llNextFreeIndex), &
        0, spvBuffer$, 0, DIR_ENTRY_SIZE )
      asFATDirEntryRamBuffer$(llNextFreeIndex) = ntos$( asFATDirEntryRamBuffer$(llNextFreeIndex), &
        DIR_IN_RAM_ADDRESS_POS, DIR_IN_RAM_ADDRESS_SIZE, lpAddr )
      lFATDirEntryRamBufSize = lFATDirEntryRamBufSize + 1
      bFATDirEntryRamBufModified = TRUE
    endif
  else
    asFATDirEntryRamBuffer$(llIndex) = stos$( asFATDirEntryRamBuffer$(llIndex), &
      0, spvBuffer$, 0, DIR_ENTRY_SIZE )
    bFATDirEntryRamBufModified = TRUE
  endif
#else
  call vHdWriteString( lpAddr, lpPos, lpSize, spvBuffer$ )
#endif
end

sub lFATFindNameInDirEntries( string spSearchedDirName$; var long lpvDirEntryAddr )
  long llIndex, llAddr, llPosInStr
  string slName$(DIR_NAME_SIZE)

#ifdef FS_SECURE_INIT
  slName$ = "" '$JK
#endif	'' FS_SECURE_INIT

  lpvDirEntryAddr = - 1
#ifdef DIR_ENTRIES_IN_RAM
  if lFATDirEntryRamBufSize > 0 then
    llIndex = 0
    while lpvDirEntryAddr = -1 and llIndex < lFATDirEntryRamBufSize
      if asFATDirEntryRamBuffer$(llIndex) <> "" then
        slName$ = mid$( asFATDirEntryRamBuffer$(llIndex), DIR_NAME_POS, DIR_NAME_SIZE )
	llPosInStr = chk_fnam( slName$, 0, DIR_NAME_SIZE, spSearchedDirName$, 0 )
        if llPosInStr >= 0 then
''        if instr( spSearchedDirName$, slName$, 0, 4 ) = 0 then
          lpvDirEntryAddr = nfroms( asFATDirEntryRamBuffer$(llIndex), DIR_IN_RAM_ADDRESS_POS, DIR_IN_RAM_ADDRESS_SIZE )
        endif
      endif
      llIndex = llIndex + 1
    endwhile
  endif
#endif
end


#ifdef DIR_ENTRIES_IN_RAM
sub vFATInitDirEntriesInRam()
  long llIndex
  for llIndex = 0 to DIR_ENTRIES_IN_RAM_NUM - 1
    asFATDirEntryRamBuffer$(llIndex) = ""
  next
  bFATDirEntryRamBufModified = FALSE
  lFATDirEntryRamBufSize = 0
end


sub vFATSynchronizeDirEntries()
  long llIndex, llAddr
  if bFATDirEntryRamBufModified = TRUE and lFATDirEntryRamBufSize > 0 then
    for llIndex = 0 to lFATDirEntryRamBufSize - 1
      if asFATDirEntryRamBuffer$(llIndex) <> "" then
        llAddr = nfroms( asFATDirEntryRamBuffer$(llIndex), DIR_IN_RAM_ADDRESS_POS, DIR_IN_RAM_ADDRESS_SIZE )
        if llAddr > 0 then
  	  call vHdWriteString( llAddr, 0, DIR_ENTRY_SIZE, asFATDirEntryRamBuffer$(llIndex) )
        endif
      endif
    next
    call vFATInitDirEntriesInRam()
  endif
end


sub lFATGetDirEntryInRamIndex( long lpAddr; var long lpvIndex; var long lpvNextFreeIndex )
  byte blWasEntrySaved
  long llIndex, llAddr
  lpvIndex = -1
  lpvNextFreeIndex = -1
  blWasEntrySaved = FALSE
  if lFATDirEntryRamBufSize > 0 then
    llIndex = 0
    while blWasEntrySaved = FALSE and llIndex < lFATDirEntryRamBufSize
      if asFATDirEntryRamBuffer$(llIndex) <> "" then
        llAddr = nfroms( asFATDirEntryRamBuffer$(llIndex), DIR_IN_RAM_ADDRESS_POS, DIR_IN_RAM_ADDRESS_SIZE )
        if llAddr = lpAddr then
          blWasEntrySaved = TRUE
          lpvIndex = llIndex
        endif
      else
        if lpvNextFreeIndex = -1 then
          lpvNextFreeIndex = llIndex
        endif
      endif
      llIndex = llIndex + 1
    endwhile
  endif
  if lpvNextFreeIndex = -1 and lFATDirEntryRamBufSize < DIR_ENTRIES_IN_RAM_NUM then
    lpvNextFreeIndex = lFATDirEntryRamBufSize
  endif
end
#endif
''
''**************************************************************************


'' The sub initializes some additional FAT vars
sub vFATInitConfig( long lpBootRecordAddr )
  lRootSectors = wBpbRootEntries
  lRootSectors = lRootSectors * DIR_ENTRY_SIZE / wBpbBytesPerSector
  lFATSectors = bBpbNumberOfFATs
  lFATSectors = lFATSectors * wBpbSectorsPerFAT
  lDataSectors = lBpbTotalSectors - ( wBpbReservedSectors + lFATSectors + lRootSectors )
  lTotalClusters = lDataSectors / bBpbSectorsPerCluster
  lFATStart = lpBootRecordAddr + wBpbReservedSectors * wBpbBytesPerSector
  lRootAreaStart = lpBootRecordAddr + ( wBpbReservedSectors + lFATSectors ) * wBpbBytesPerSector
  lDataAreaStart = lpBootRecordAddr + ( wBpbReservedSectors + lFATSectors + lRootSectors ) * wBpbBytesPerSector
  lBytesPerCluster = wBpbBytesPerSector
  lBytesPerCluster = lBytesPerCluster * bBpbSectorsPerCluster
  lBytesPerFAT = wBpbSectorsPerFAT
  lBytesPerFAT = lBytesPerFAT * wBpbBytesPerSector
  if lTotalClusters <= FAT12_MAX_CLUSTERs then
    bFATType = FAT12
    bFATClusterPosMult = FAT12_MULT
    bFATClusterPosDiv = FAT12_DIV
  else
    if lTotalClusters <= FAT16_MAX_CLUSTERs then
      bFATType = FAT16
      bFATClusterPosMult = FAT16_MULT
      bFATClusterPosDiv = FAT16_DIV
    else
      bFATType = FAT32
      bFATClusterPosMult = FAT32_MULT
      bFATClusterPosDiv = FAT32_DIV
    endif
  endif

  alFATEndClusterMinId( FAT12 ) = FAT12_EOF
  alFATEndClusterMinId( FAT16 ) = FAT16_EOF
  alFATEndClusterMinId( FAT32 ) = FAT32_EOF

  alFATEndClusterMaxId( FAT12 ) = FAT12_EOF_MAX
  alFATEndClusterMaxId( FAT16 ) = FAT16_EOF_MAX
  alFATEndClusterMaxId( FAT32 ) = FAT32_EOF_MAX

  alFATInvalidClusterId( FAT12 ) = FAT12_BAD
  alFATInvalidClusterId( FAT16 ) = FAT16_BAD
  alFATInvalidClusterId( FAT32 ) = FAT32_BAD

  bFATFmtWinFlag = FALSE
end


sub lFATGetBootSectAddr( var long lpvValidAddrLog; var byte bpvIsFound )
  byte bpReadOk
  long llPageId, llNextValidAddr, llNextPageIdAddr, llCurBlockLim
  string slFSName$(BS_FILE_SYSTEM_SIZE)

#ifdef FS_SECURE_INIT
  slFSName$ = "" '$JK
#endif	'' FS_SECURE_INIT

  bpvIsFound = FALSE
  bpReadOk = TRUE
  llPageId = 0
  llNextValidAddr = 0
  while bpReadOk = TRUE
    llCurBlockLim = llNextValidAddr + lHdBytesInBlock
    while bpReadOk = TRUE and llNextValidAddr < llCurBlockLim
      llNextPageIdAddr = llNextValidAddr + BOOT_MAGIC_POS
      call lHdReadNum( llNextPageIdAddr, WORDSIZE, llPageId, bpReadOk )
      if bpReadOk = TRUE and llPageId = BOOT_MAGIC then
        llNextPageIdAddr = llNextValidAddr + BS_BPB_POS
        call lHdReadNum( llNextPageIdAddr, WORDSIZE, llPageId, bpReadOk )
        if bpReadOk = TRUE and llPageId <> 0 then
          llNextPageIdAddr = llNextValidAddr + BS_FILE_SYSTEM_POS
          call sHdReadString( llNextPageIdAddr, BS_FILE_SYSTEM_SIZE, slFSName$, bpReadOk )
          if bpReadOk = TRUE then
            slFSName$ = remove$( slFSName$, " " )
            if slFSName$ = FAT12_NAME or slFSName$ = FAT16_NAME then
              lpvValidAddrLog = llNextValidAddr
              bpvIsFound = TRUE
              return
            endif
          endif
        endif
      endif
      llNextValidAddr = llNextValidAddr + lHdBytesInPage
    endwhile
  endwhile
end


'' The sub reads boot record from the address "lpBootRecordAddr" and saves the fields
'' of boot record in global vars; additional FAT vars are initialized
'' in the sub "vFATInitConfig" called from here in good time
sub vFATGetBootRecord( long lpBootRecordAddrLog )
  byte blReadOk, blBootRecNotEmpty
  string slBootRecord$( BOOT_RECORD_SIZE )

  slBootRecord$ = ""
  call sHdReadString( lpBootRecordAddrLog, BOOT_RECORD_SIZE, slBootRecord$, blReadOk )

  if slBootRecord$ <> "" then
    blBootRecNotEmpty = TRUE
  else
    blBootRecNotEmpty = FALSE
  endif

  if blReadOk = TRUE and blBootRecNotEmpty = TRUE then
    '' Read the beginning boot record fields, that are common to all FATs
    sBsJmp$ = mid$( slBootRecord$, BS_JMP_POS, BS_JMP_SIZE )
    sBsOemName$ = mid$( slBootRecord$, BS_OEM_NAME_POS, BS_OEM_NAME_SIZE )

    wBpbBytesPerSector = nfroms( slBootRecord$, BPB_BYTES_PER_SECT_POS, BPB_BYTES_PER_SECT_SIZE )
    bBpbSectorsPerCluster = nfroms( slBootRecord$, BPB_SECT_PER_CLUSTER_POS, BPB_SECT_PER_CLUSTER_SIZE )
    wBpbReservedSectors = nfroms( slBootRecord$, BPB_RESERVED_SECT_POS, BPB_RESERVED_SECT_SIZE )
    bBpbNumberOfFATs = nfroms( slBootRecord$, BPB_NUMBER_OF_FATS_POS, BPB_NUMBER_OF_FATS_SIZE )
    wBpbRootEntries = nfroms( slBootRecord$, BPB_ROOT_ENTRIES_POS, BPB_ROOT_ENTRIES_SIZE )
    lBpbTotalSectors = nfroms( slBootRecord$, BPB_TOTAL_SECT_POS, BPB_TOTAL_SECT_SIZE )
    bBpbMedia = nfroms( slBootRecord$, BPB_MEDIA_POS, BPB_MEDIA_SIZE )
    wBpbSectorsPerFAT = nfroms( slBootRecord$, BPB_SECT_PER_FAT_POS, BPB_SECT_PER_FAT_SIZE )
    wBpbSectorsPerTrack = nfroms( slBootRecord$, BPB_SECT_PER_TRACK_POS, BPB_SECT_PER_TRACK_SIZE )
    wBpbHeadsPerCylinder = nfroms( slBootRecord$, BPB_HEADS_PER_CYL_POS, BPB_HEADS_PER_CYL_SIZE )
    lBpbHiddenSectors = nfroms( slBootRecord$, BPB_HIDDEN_SECT_POS, BPB_HIDDEN_SECT_SIZE )
    lBpbTotalSectorsBig = nfroms( slBootRecord$, BPB_TOTAL_SECT_BIG_POS, BPB_TOTAL_SECT_BIG_SIZE )

    if lBpbTotalSectors = 0 then
      lBpbTotalSectors = lBpbTotalSectorsBig - lBpbHiddenSectors
    endif

    '' Now we have information enough to decide which type of FAT is meant
    call vFATInitConfig( lpBootRecordAddrLog )

    '' The remained parts have the different offsets in the boot record for FAT32 and for others
    if bFATType = FAT32 then
      lBpb32SectorsPerFAT = nfroms( slBootRecord$, BPB32_SECT_PER_FAT_POS, BPB32_SECT_PER_FAT_SIZE )
      wBpb32Flags = nfroms( slBootRecord$, BPB32_FLAGS_POS, BPB32_FLAGS_SIZE )
      wBpb32Version = nfroms( slBootRecord$, BPB32_VERSION_POS, BPB32_VERSION_SIZE )
      lBpb32RootCluster = nfroms( slBootRecord$, BPB32_ROOT_CLUSTER_POS, BPB32_ROOT_CLUSTER_SIZE )
      wBpb32InfoSector = nfroms( slBootRecord$, BPB32_INFO_SECT_POS, BPB32_INFO_SECT_SIZE )
      wBpb32BootBackupStart = nfroms( slBootRecord$, BPB32_BOOT_BACKUP_START_POS, BPB32_BOOT_BACKUP_START_SIZE )
      abBpb32Reserved$ = mid$( slBootRecord$, BPB32_RESERVED_POS, BPB32_RESERVED_SIZE )

      bBsDriveNumber = nfroms( slBootRecord$, BS32_DRIVE_NUMBER_POS, BS32_DRIVE_NUMBER_SIZE )
      bBsUnused = nfroms( slBootRecord$, BS32_UNUSED_POS, BS32_UNUSED_SIZE )
      bBsExtBootSignature = nfroms( slBootRecord$, BS32_EXT_BOOT_SIGN_POS, BS32_EXT_BOOT_SIGN_SIZE )
      lBsSerialNumber = nfroms( slBootRecord$, BS32_SERIAL_NUMBER_POS, BS32_SERIAL_NUMBER_SIZE )
      sBsVolumeLabel$ = mid$( slBootRecord$, BS32_VOLUME_LABEL_POS, BS32_VOLUME_LABEL_SIZE )
      sBsFileSystem$ = mid$( slBootRecord$, BS32_FILE_SYSTEM_POS, BS32_FILE_SYSTEM_SIZE )
      sBsBootCode$ = mid$( slBootRecord$, BS32_BOOT_CODE_POS, BS32_BOOT_CODE_SIZE )
    else
      bBsDriveNumber = nfroms( slBootRecord$, BS_DRIVE_NUMBER_POS, BS_DRIVE_NUMBER_SIZE )
      bBsUnused = nfroms( slBootRecord$, BS_UNUSED_POS, BS_UNUSED_SIZE )
      bBsExtBootSignature = nfroms( slBootRecord$, BS_EXT_BOOT_SIGN_POS, BS_EXT_BOOT_SIGN_SIZE )
      lBsSerialNumber = nfroms( slBootRecord$, BS_SERIAL_NUMBER_POS, BS_SERIAL_NUMBER_SIZE )
      sBsVolumeLabel$ = mid$( slBootRecord$, BS_VOLUME_LABEL_POS, BS_VOLUME_LABEL_SIZE )
      sBsFileSystem$ = mid$( slBootRecord$, BS_FILE_SYSTEM_POS, BS_FILE_SYSTEM_SIZE )
      sBsBootCode$ = mid$( slBootRecord$, BS_BOOT_CODE_POS, BS_BOOT_CODE_SIZE )
    endif
  endif
end


sub bFATReadBootRecordToString( var string spvBootRecord$; var byte bpvIsBootRecRead )
  byte blFATBootSectorFound, blReadOk
  long llFATBootSectAddrLog

  bpvIsBootRecread = FALSE
  if max_len( spvBootRecord$ ) < BOOT_RECORD_SIZE then
    return
  endif
  llFATBootSectAddrLog = 0h
  call lFATGetBootSectAddr( llFATBootSectAddrLog, blFATBootSectorFound )

  if blFATBootSectorFound = TRUE then
    call sHdReadString( llFATBootSectAddrLog, BOOT_RECORD_SIZE, spvBootRecord$, blReadOk )
    if blReadOk = TRUE then
      if spvBootRecord$ <> "" then
        bpvIsBootRecRead = TRUE
      endif
    endif
  endif
end


'' The sub reads a dir entry from the address "lpAddr" and saves the fields
'' of the dir entry in globals
sub bFATGetDirEntry( long lpAddr; var byte bpvIsAvailable )
  byte blIsOk
  string slDirEntry$( DIR_ENTRY_SIZE )

  slDirEntry$ = ""
''''  call sHdReadString( lpAddr, DIR_ENTRY_SIZE, slDirEntry$, blIsOk )
  call sFATReadDirEntryString( lpAddr, DIR_ENTRY_SIZE, slDirEntry$, blIsOk )

  if slDirEntry$ <> "" then
    sDirName$ = mid$( slDirEntry$, DIR_NAME_POS, DIR_NAME_SIZE )
    bDirAttr = nfroms( slDirEntry$, DIR_ATTR_POS, BYTESIZE )
    bDirReserved = nfroms( slDirEntry$, DIR_RESERVED_POS, BYTESIZE )
    bDirCreate10MS = nfroms( slDirEntry$, DIR_CREATE10MS_POS, BYTESIZE )
    wDirCreateTime = nfroms( slDirEntry$, DIR_CREATETIME_POS, WORDSIZE )
    wDirCreateDate = nfroms( slDirEntry$, DIR_CREATEDATE_POS, WORDSIZE )
    wDirAccessDate = nfroms( slDirEntry$, DIR_ACCESSDATE_POS, WORDSIZE )
    wDirClusterHigh = nfroms( slDirEntry$, DIR_CLUSTERHIGH_POS, WORDSIZE )
    wDirTime = nfroms( slDirEntry$, DIR_TIME_POS, WORDSIZE )
    wDirDate = nfroms( slDirEntry$, DIR_DATE_POS, WORDSIZE )
    wDirCluster = nfroms( slDirEntry$, DIR_CLUSTER_POS, WORDSIZE )
    lDirSize = nfroms( slDirEntry$, DIR_SIZE_POS, LONGSIZE )

    bpvIsAvailable = nfroms( slDirEntry$, DIR_SPEC_INFO_POS, BYTESIZE )
    if bpvIsAvailable = DIR_NOT_USED_ID or bpvIsAvailable = DIR_ERASED_ID then
      bpvIsAvailable = FALSE
    else
      bpvIsAvailable = TRUE
    endif
  endif
end


'' The sub writes a dir entry from globalg to the address "lpAddr"
sub bFATWriteDirEntry( long lpAddr )
  string slDirEntry$( DIR_ENTRY_SIZE )

  slDirEntry$ = ""
  slDirEntry$ = stos$( slDirEntry$, DIR_NAME_POS, sDirName$, DIR_NAME_SIZE )
  slDirEntry$ = ntos$( slDirEntry$, DIR_ATTR_POS, BYTESIZE, bDirAttr )
  slDirEntry$ = ntos$( slDirEntry$, DIR_RESERVED_POS, BYTESIZE, bDirReserved )
  slDirEntry$ = ntos$( slDirEntry$, DIR_CREATE10MS_POS, BYTESIZE, bDirCreate10MS )
  slDirEntry$ = ntos$( slDirEntry$, DIR_CREATETIME_POS, WORDSIZE, wDirCreateTime )
  slDirEntry$ = ntos$( slDirEntry$, DIR_CREATEDATE_POS, WORDSIZE, wDirCreateDate )
  slDirEntry$ = ntos$( slDirEntry$, DIR_ACCESSDATE_POS, WORDSIZE, wDirAccessDate )
  slDirEntry$ = ntos$( slDirEntry$, DIR_CLUSTERHIGH_POS, WORDSIZE, wDirClusterHigh )
  slDirEntry$ = ntos$( slDirEntry$, DIR_TIME_POS, WORDSIZE, wDirTime )
  slDirEntry$ = ntos$( slDirEntry$, DIR_DATE_POS, WORDSIZE, wDirDate )
  slDirEntry$ = ntos$( slDirEntry$, DIR_CLUSTER_POS, WORDSIZE, wDirCluster )
  slDirEntry$ = ntos$( slDirEntry$, DIR_SIZE_POS, LONGSIZE, lDirSize )

''''  call vHdWriteString( lpAddr, 0, DIR_ENTRY_SIZE, slDirEntry$ )
  call vFATWriteDirEntryString( lpAddr, 0, DIR_ENTRY_SIZE, slDirEntry$ )

end


'' The sub reads a long name file dir entry from the address "lpAddr" and
'' saves the fields of the lfn dir entry in globals
sub bFATGetDirEntryLFN( long lpAddr; var byte bpvIsAvailable )
  byte blIsOk
  string slDirEntry$( DIR_ENTRY_SIZE )

  slDirEntry$ = ""
''''  call sHdReadString( lpAddr, DIR_ENTRY_SIZE, slDirEntry$, blIsOk )
  call sFATReadDirEntryString( lpAddr, DIR_ENTRY_SIZE, slDirEntry$, blIsOk )

  if slDirEntry$ <> "" then
    bLfnSequence = nfroms( slDirEntry$, LFN_SEQUENCE_POS, BYTESIZE )
    sLfnName1$ = mid$( slDirEntry$, LFN_NAME1_POS, 10 )
    bLfnAttr = nfroms( slDirEntry$, LFN_ATTR_POS, BYTESIZE )
    bLfnReserved = nfroms( slDirEntry$, LFN_RESERVED_POS, BYTESIZE )
    bLfnChecksum = nfroms( slDirEntry$, LFN_CHECKSUM_POS, BYTESIZE )
    sLfnName2$ = mid$( slDirEntry$, LFN_NAME2_POS, 12 )
    wLfnCluster = nfroms( slDirEntry$, LFN_CLUSTER_POS, WORDSIZE )
    sLfnName3$ = mid$( slDirEntry$, LFN_NAME3_POS, 4 )

    bpvIsAvailable = nfroms( slDirEntry$, DIR_SPEC_INFO_POS, BYTESIZE )
    if bpvIsAvailable = DIR_NOT_USED_ID or bpvIsAvailable = DIR_ERASED_ID then
      bpvIsAvailable = FALSE
    else
      bpvIsAvailable = TRUE
    endif
  endif
end


sub bFATGetLongNameLFN( long lpAliasDirEntryAddr; var string spvLongFileName$; &
  long lpMaxNameSize; var byte bpvExist )
  byte blIsAvailable, blLFNAttr
  long llCurDirEntryAddr
  bpvExist = FALSE
  spvLongFileName$ = ""
  blLFNAttr = LFN_TAG_ATTR
  llCurDirEntryAddr = lpAliasDirEntryAddr - DIR_ENTRY_SIZE
  while blLFNAttr = LFN_TAG_ATTR
    call bFATGetDirEntryLFN( llCurDirEntryAddr, blIsAvailable )
    if blIsAvailable = TRUE and bLfnAttr = LFN_TAG_ATTR and bLfnSequence <> DIR_ERASED_ID then
      blLFNAttr = bLfnAttr
      if len(spvLongFileName$) + len(sLfnName1$) < lpMaxNameSize then
        spvLongFileName$ = spvLongFileName$ + sLfnName1$
      else
        blLFNAttr = 0
      endif
      if len(spvLongFileName$) + len(sLfnName2$) < lpMaxNameSize then
        spvLongFileName$ = spvLongFileName$ + sLfnName2$
      else
        blLFNAttr = 0
      endif
      if len(spvLongFileName$) + len(sLfnName3$) < lpMaxNameSize then
        spvLongFileName$ = spvLongFileName$ + sLfnName3$
      else
        blLFNAttr = 0
      endif
      llCurDirEntryAddr = llCurDirEntryAddr - DIR_ENTRY_SIZE
    else
      blLFNAttr = 0
    endif
  endwhile
  if len(spvLongFileName$) > 0 then
    bpvExist = TRUE
  endif
end


'' The sub builds a long name dir entry and writes it to the address "lpAddr"
sub bFATWriteDirEntryLFN( string spLongName$; word wpSlotIndex; &
  byte bpAliasCheckSum; long lpAddr; var byte bpvIsEndSlot )

  string slDirEntry$( DIR_ENTRY_SIZE )
  string slPartName$( LFN_LETTERS_PER_SLOT )
  word wlPartIndex, wlPartNameLen, wlIndex, wlI

#ifdef FS_SECURE_INIT
  slDirEntry$ = "" '$JK
  slPartName$ = "" '$JK
#endif	'' FS_SECURE_INIT

  bpvIsEndSlot = FALSE
  slDirEntry$ = fill$( "FF"%, DIR_ENTRY_SIZE )
  wlPartIndex = ( wpSlotIndex - 1 ) * LFN_LETTERS_PER_SLOT
  wlPartNameLen = len( spLongName$ ) - wlPartIndex
  if wlPartNameLen > LFN_LETTERS_PER_SLOT then
    wlPartNameLen = LFN_LETTERS_PER_SLOT
  endif
  slPartName$ = mid$( spLongName$, wlPartIndex, wlPartNameLen )
  if wlPartNameLen < LFN_LETTERS_PER_SLOT then
    slPartName$ = slPartName$ + "00"%
    wlPartNameLen = wlPartNameLen + 1
    bpvIsEndSlot = TRUE
  endif

  if bpvIsEndSlot = TRUE then
    slDirEntry$ = ntos$( slDirEntry$, LFN_SEQUENCE_POS, BYTESIZE, wpSlotIndex bitor LFN_SEQ_LAST )
  else
    slDirEntry$ = ntos$( slDirEntry$, LFN_SEQUENCE_POS, BYTESIZE, wpSlotIndex )
  endif
  wlI = 0
  wlIndex = 0
  while wlIndex < 5 and wlIndex < wlPartNameLen
    slDirEntry$ = stos$( slDirEntry$, LFN_NAME1_POS + wlI, slPartName$, wlIndex, BYTESIZE )
    slDirEntry$ = ntos$( slDirEntry$, LFN_NAME1_POS + wlI + 1, BYTESIZE, 0 )
    wlI = wlI + 2
    wlIndex = wlIndex + 1
  endwhile
  slDirEntry$ = ntos$( slDirEntry$, LFN_ATTR_POS, BYTESIZE, LFN_TAG_ATTR )
  slDirEntry$ = ntos$( slDirEntry$, LFN_RESERVED_POS, BYTESIZE, 0 )
  slDirEntry$ = ntos$( slDirEntry$, LFN_CHECKSUM_POS, BYTESIZE, bpAliasCheckSum )
  wlI = 0
  while wlIndex < 11 and wlIndex < wlPartNameLen
    slDirEntry$ = stos$( slDirEntry$, LFN_NAME2_POS + wlI, slPartName$, wlIndex, BYTESIZE )
    slDirEntry$ = ntos$( slDirEntry$, LFN_NAME2_POS + wlI + 1, BYTESIZE, 0 )
    wlI = wlI + 2
    wlIndex = wlIndex + 1
  endwhile
  slDirEntry$ = ntos$( slDirEntry$, LFN_CLUSTER_POS, WORDSIZE, 0 )
  wlI = 0
  while wlIndex < 13 and wlIndex < wlPartNameLen
    slDirEntry$ = stos$( slDirEntry$, LFN_NAME3_POS + wlI, slPartName$, wlIndex, BYTESIZE )
    slDirEntry$ = ntos$( slDirEntry$, LFN_NAME3_POS + wlI + 1, BYTESIZE, 0 )
    wlI = wlI + 2
    wlIndex = wlIndex + 1
  endwhile

''''  call vHdWriteString( lpAddr, 0, DIR_ENTRY_SIZE, slDirEntry$ )
  call vFATWriteDirEntryString( lpAddr, 0, DIR_ENTRY_SIZE, slDirEntry$ )
end


sub bFATCalcAliasCheckSum( string spAliasName$; var byte bpvCheckSum )
  byte blI
  bpvCheckSum = 0
  for blI = 0 to 10
    bpvCheckSum = ( ( ( bpvCheckSum bitand 1 ) shl 7 ) bitor &
      ( ( bpvCheckSum bitand 0FEh ) shr 1 ) ) + nfroms( spAliasName$, blI, 1 )
  next
end


sub vFATWriteSlotsLFN( string spLongName$; string spAliasName$; var long lpvAddr )
  byte blCheckSum, blIsEndSlot
  word wlSlotIndex
  long llSlotNumber
  call bFATCalcAliasCheckSum( spAliasName$, blCheckSum )
  blIsEndSlot = FALSE
  wlSlotIndex = 1
  '' calculate number of slots for long file name
  call lCalcNumOfSlotsLFN( spLongName$, llSlotNumber )
  lpvAddr = lpvAddr + ( ( llSlotNumber - 1 ) * DIR_ENTRY_SIZE )
  while blIsEndSlot = FALSE
    call bFATWriteDirEntryLFN( spLongName$, wlSlotIndex, blCheckSum, &
      lpvAddr, blIsEndSlot )
    wlSlotIndex = wlSlotIndex + 1
    lpvAddr = lpvAddr - DIR_ENTRY_SIZE
  endwhile
  '' set the address for alias dir entry
  lpvAddr = lpvAddr + ( ( llSlotNumber + 1 ) * DIR_ENTRY_SIZE )
end


sub lCalcNumOfSlotsLFN( string spLongName$; var long lpvSlots )
  lpvSlots = len( spLongName$ ) / LFN_LETTERS_PER_SLOT
  if mod( len( spLongName$ ), LFN_LETTERS_PER_SLOT ) > 0 then
    lpvSlots = lpvSlots + 1
  endif
end


'' The sub searches for the name "spSearchedDirName$" in the area
'' beginning with "lpStartAddr" up to the end of the current cluster
'' lpvDirEntryAddr=-2 if the name was not found and the end of the cluster was reached
'' lpvDirEntryAddr>=0 if the name was found; lpvDirEntryAddr contains the address of
''   the fields describing the directory or the file; those fields are assigned to some
''   global vars in the sub "bFATGetDirEntry"
sub lFATFindDirEntryInCluster( string spSearchedDirName$; long lpStartAddr; var long lpvDirEntryAddr )
  byte blIsAvailable, blEndOfSearch, blIsSpecDirEntry
  long llCurAddr

  llCurAddr = lpStartAddr
  if mod( llCurAddr, DIR_ENTRY_SIZE ) <> 0 then
    llCurAddr = llCurAddr + ( DIR_ENTRY_SIZE - mod( llCurAddr, DIR_ENTRY_SIZE ) )
  endif
  call lFATFindNameInDirEntries( spSearchedDirName$, lpvDirEntryAddr )
  if lpvDirEntryAddr < 0 then
    blEndOfSearch = FALSE
    while blEndOfSearch = FALSE
      call lHdFindStringInMemBlock( spSearchedDirName$, llCurAddr, lBytesPerCluster, DIR_ENTRY_SIZE, lpvDirEntryAddr )
      blEndOfSearch = TRUE
      if lpvDirEntryAddr >= 0 then
        if mod( lpvDirEntryAddr, DIR_ENTRY_SIZE ) <> 0 then
    	  lpvDirEntryAddr = -2
          blEndOfSearch = TRUE
        else
    	  call bFATGetDirEntry( lpvDirEntryAddr, blIsAvailable )
    	  if blIsAvailable = FALSE then
      	    lpvDirEntryAddr = -2
          else
            blIsSpecDirEntry = FALSE
            if trim$( sDirName$, " ", -1 ) = "." or trim$( sDirName$, " ", -1 ) = ".." then
              blIsSpecDirEntry = TRUE
            endif
            if bDirAttr = LFN_TAG_ATTR or blIsSpecDirEntry = TRUE then
              llCurAddr = lpvDirEntryAddr + DIR_ENTRY_SIZE
              blEndOfSearch = FALSE
            endif
    	  endif
        endif
      endif
    endwhile
  endif
end


sub vFATMarkAsErasedSlotsLFN( long lpAliasDirEntryAddr )
  byte blIsAvailable, blLFNAttr
  long llCurDirEntryAddr
  string slErasedMark$(1)
  slErasedMark$ = DIR_ERASED_ID_STR
  blLFNAttr = LFN_TAG_ATTR
  llCurDirEntryAddr = lpAliasDirEntryAddr - DIR_ENTRY_SIZE
  while blLFNAttr = LFN_TAG_ATTR
    call bFATGetDirEntryLFN( llCurDirEntryAddr, blIsAvailable )
    if blIsAvailable = TRUE and bLfnAttr = LFN_TAG_ATTR then
      blLFNAttr = bLfnAttr
      call vHdWriteString( llCurDirEntryAddr, 0, 1, slErasedMark$ )
      llCurDirEntryAddr = llCurDirEntryAddr - DIR_ENTRY_SIZE
    else
      blLFNAttr = 0
    endif
  endwhile
end


sub vFATMarkDirEntryAsErased( long lpDirEntryAddr )
  string slErasedMark$(1)
  slErasedMark$ = DIR_ERASED_ID_STR
  call vHdWriteString( lpDirEntryAddr, 0, 1, slErasedMark$ )
  call vFATMarkAsErasedSlotsLFN( lpDirEntryAddr )
end


sub vFATMarkSpecDirEntriesAsErased( long lpDirAddr )
  long llEntryAddr
  string slErasedMark$(1)
  slErasedMark$ = DIR_ERASED_ID_STR
  llEntryAddr = lpDirAddr
  call vHdWriteString( llEntryAddr, 0, 1, slErasedMark$ )
  llEntryAddr = llEntryAddr + DIR_ENTRY_SIZE
  call vHdWriteString( llEntryAddr, 0, 1, slErasedMark$ )
end


sub bFATIsDirEmpty( word wpFirstDirCluster; var byte bpvIsDirEmpty )
  byte blIsEntryEmpty, blIsEnd
  long llAddrCounter, llCurCluster, llCurAddr, llNextCluster
  llAddrCounter = 0
  llCurCluster = wpFirstDirCluster
  llCurAddr = -1
  blIsEntryEmpty = TRUE
  blIsEnd = FALSE
  while blIsEntryEmpty = TRUE and blIsEnd = FALSE
    call lFATClusterToSectorAddr( llCurCluster, llCurAddr )
    llAddrCounter = llCurAddr
    while blIsEntryEmpty = TRUE and llAddrCounter < llCurAddr + lBytesPerCluster
      call bFATIsDirEntryForDel( llAddrCounter, blIsEntryEmpty )
      llAddrCounter = llAddrCounter + DIR_ENTRY_SIZE
    endwhile
    if blIsEntryEmpty = TRUE then
      call lFATGetNextClusterAddr( llCurCluster, 1, llNextCluster )
      llCurCluster = llNextCluster
      if llCurCluster < 1 then
        blIsEnd = TRUE
      else
        call bFATIsEndCluster( llCurCluster, blIsEnd )
      endif
    endif
  endwhile
  bpvIsDirEmpty = blIsEntryEmpty
end


'' The sub checks whether the dir entry is empty for deletion. It means:
'' the entry has 0 or e5 in the very first byte, additionally
'' the entries with dir names "." and ".." will be qualified as empty,
'' because to delete a directory we should guarantee that it is empty of files or
'' or of other directories, but the "." and ".." entries are service entries only
sub bFATIsDirEntryForDel( long lpDirEntryAddr; var byte bpvIsDirForDel )
  byte blIsAvailable
  bpvIsDirForDel = FALSE
  call bFATGetDirEntry( lpDirEntryAddr, blIsAvailable )
  if blIsAvailable = FALSE then
    bpvIsDirForDel = TRUE
  else
    if remove$( sDirName$, " " ) = "." or remove$( sDirName$, " " ) = ".." then
      bpvIsDirForDel = TRUE
    endif
  endif
end


'' The sub returns the id of the next cluster in cluster chain of FAT;
'' If next cluster cannot be found, the return value "lpvNextClusterAddr" is -1
sub lFATGetNextClusterAddr( long lpCurClusterAddr, lpRepeatNum; var long lpvNextClusterAddr )
  long llPhysAddr
  byte blIsValid, blIsEnd, blIsOk

#ifdef FAT_IN_RAM
  call lFATTableGetNextClusterAddr( lpCurClusterAddr, lpRepeatNum, lpvNextClusterAddr )
#else
  call bFATIsEndCluster( lpCurClusterAddr, blIsEnd )
  if blIsEnd = TRUE then
    lpvNextClusterAddr = -1
    return
  endif

  llPhysAddr = lFATStart + lpCurClusterAddr * bFATClusterPosMult / bFATClusterPosDiv
''''  call lHdReadNum( llPhysAddr, 4, lpvNextClusterAddr, blIsOk )
  call lFATReadTableNum( llPhysAddr, 4, lpvNextClusterAddr, blIsOk )
  switch bFATType
    case FAT12:
      lpvNextClusterAddr = lpvNextClusterAddr bitand 0ffffh
      if mod( lpCurClusterAddr, 2 ) = 0 then
        lpvNextClusterAddr = lpvNextClusterAddr bitand 0fffh
      else
        lpvNextClusterAddr = ( lpvNextClusterAddr shr 4 ) bitand 0fffh
      endif
    case FAT16:
      lpvNextClusterAddr = lpvNextClusterAddr bitand 0ffffh
    case FAT32:
  endswitch

  call bFATIsClusterValid( lpCurClusterAddr, blIsValid )
  if blIsValid = FALSE then
    lpvNextClusterAddr = -1
  endif
#endif
end


'' The sub returns the address of the last cluster in the cluster chain
sub lFATGetLastClusterInChain( long lpCurClusterAddr; var long lpvLastClusterAddr )
  byte blIsEnd
  long llCurCluster, llNextCluster
  llCurCluster = lpCurClusterAddr
  call bFATIsEndCluster( llCurCluster, blIsEnd )
  while blIsEnd = FALSE
    call lFATGetNextClusterAddr( llCurCluster, 1, llNextCluster )
    if llNextCluster < 1 then
      blIsEnd = TRUE
    else
      call bFATIsEndCluster( llNextCluster, blIsEnd )
      if blIsEnd = FALSE then
        llCurCluster = llNextCluster
      endif
    endif
  endwhile
  if llCurCluster < 1 then
    lpvLastClusterAddr = lpCurClusterAddr
  else
    lpvLastClusterAddr = llCurCluster
  endif
end


'' The sub checks, whether the current cluster is the valid one
sub bFATIsClusterValid( long lpCluster; var byte bpvIsValid )
  bpvIsValid = TRUE
  if lpCluster = 1 or lpCluster = alFATInvalidClusterId( bFATType ) then
    bpvIsValid = FALSE
  endif
end


'' The sub checks, whether the current cluster is the end cluster
'' in cluster chain of FAT for this directory or file
sub bFATIsEndCluster( long lpCluster; var byte bpvIsEnd )
  bpvIsEnd = FALSE
  if lpCluster >= alFATEndClusterMinId( bFATType ) and &
    lpCluster <= alFATEndClusterMaxId( bFATType ) then
      bpvIsEnd = TRUE
  endif
end


'' The sub makes an address (in bytes) from a cluster id
sub lFATClusterToSectorAddr( long lpClusterAddr; var long lpvSectorAddr )
  lpvSectorAddr = ( lpClusterAddr - 2 ) * lBytesPerCluster + lDataAreaStart
end


'' The sub makes a block id from a cluster id
sub lFATClusterToBlockId( long lpClusterAddr; var long lpvBlockId )
  lpvBlockId = ( ( lpClusterAddr - 2 ) * lBytesPerCluster ) / lHdBytesInBlock
end


'' The sub writes a new value to the cluster cell address in FAT
sub lFATWriteCluster( long lpFatClusterAddr, lpNewClusterValue )
  long llAddrInFAT, llAddrInFAT2, llNumBytesToBeWritten, llPrevValue, llNewValue
  byte blIsValid, blIsOk
  string slBuffer$(8)

#ifdef FS_SECURE_INIT
  slBuffer$ = "" '$JK
#endif	'' FS_SECURE_INIT

  llAddrInFAT = lFATStart + lpFatClusterAddr * bFATClusterPosMult / bFATClusterPosDiv
  llAddrInFAT2 = llAddrInFAT + lBytesPerFAT
  llNumBytesToBeWritten = 0
  llNewValue = lpNewClusterValue
  switch bFATType
    case FAT12:
      llNewValue = llNewValue bitand 0fffh
''''      call lHdReadNum( llAddrInFAT, 2, llPrevValue, blIsOk )
      call lFATReadTableNum( llAddrInFAT, 2, llPrevValue, blIsOk )
      if mod( lpFatClusterAddr, 2 ) = 0 then
        llNewValue = llNewValue bitor ( llPrevValue bitand 0f000h )
      else
        llNewValue = ( llNewValue shl 4 ) bitor ( llPrevValue bitand 0fh )
      endif
      llNumBytesToBeWritten = 2
    case FAT16:
      llNumBytesToBeWritten = 2
    case FAT32:
      llNumBytesToBeWritten = 4
  endswitch

  slBuffer$ = ""
  slBuffer$ = ntos$( slBuffer$, 0, llNumBytesToBeWritten, llNewValue )
''''  call vHdWriteString( llAddrInFAT, 0, llNumBytesToBeWritten, slBuffer$ )
''''  '' update the copy of FAT also
''''  call vHdWriteString( llAddrInFAT2, 0, llNumBytesToBeWritten, slBuffer$ )
  call vFATWriteTableString( llAddrInFAT, 0, llNumBytesToBeWritten, slBuffer$ )
end


sub lFATGetNextFreeCluster( long lpCurClusterAddr; var long lpvFreeClusterAddr )
  byte blIsRead, blIsFound, blIsAddrOk
  long llAddrInFatChain, llAddrCounter, llClusterCounter, llClustersInBuf, llIndx
  string slBuffer$(READ_FAT_BUFFER_SIZE)

#ifdef FS_SECURE_INIT
  slBuffer$ = "" '$JK
#endif	'' FS_SECURE_INIT

  lpvFreeClusterAddr = -1
  blIsFound = FALSE
  llClusterCounter = lpCurClusterAddr
  if bFATType = FAT12 and llClusterCounter = 0 then
    llClusterCounter = 2
  endif
  llAddrCounter = llClusterCounter * bFATClusterPosMult / bFATClusterPosDiv
  while blIsFound = FALSE and llAddrCounter < lBytesPerFAT
    llAddrInFatChain = lFATStart + llAddrCounter
    call sFATReadTableString( llAddrInFatChain, READ_FAT_BUFFER_SIZE, slBuffer$, blIsRead )
''''    call sHdReadString( llAddrInFatChain, READ_FAT_BUFFER_SIZE, slBuffer$, blIsRead )
    llIndx = 0
    blIsAddrOk = FALSE
    while blIsAddrOk = FALSE and llIndx < READ_FAT_BUFFER_SIZE
      blIsAddrOk = TRUE
      switch bFATType
        case FAT12:
          lpvFreeClusterAddr = instr( "<00>", slBuffer$, llIndx, 4 )
      	  if lpvFreeClusterAddr <> -1 then
            if mod( ( lpvFreeClusterAddr + llAddrCounter ) * bFATClusterPosDiv / &
              bFATClusterPosMult, 2 ) = 0 then
                if nfroms( slBuffer$, lpvFreeClusterAddr + 1, 1 ) bitand 0fh <> 0 then
                  blIsAddrOk = FALSE
                endif
      	    else
              if nfroms( slBuffer$, lpvFreeClusterAddr - 1, 1 ) bitand 0f0h <> 0 then
                blIsAddrOk = FALSE
              endif
      	    endif
          else
            lpvFreeClusterAddr = READ_FAT_BUFFER_SIZE - 1
            blIsAddrOk = FALSE
          endif
        case FAT16:
          lpvFreeClusterAddr = instr( "<00><00>", slBuffer$, llIndx, 4 )
    	  if lpvFreeClusterAddr < 0 then
            blIsAddrOk = FALSE
            lpvFreeClusterAddr = READ_FAT_BUFFER_SIZE - 1
          else
            if mod( lpvFreeClusterAddr + llAddrCounter, 2 ) <> 0 then
              blIsAddrOk = FALSE
            endif
    	  endif
        case FAT32:
          lpvFreeClusterAddr = instr( "<00><00><00><00>", slBuffer$, llIndx, 4 )
    	  if lpvFreeClusterAddr < 0 then
            blIsAddrOk = FALSE
            lpvFreeClusterAddr = READ_FAT_BUFFER_SIZE - 1
    	  endif
      endswitch
      if blIsAddrOk = FALSE then
        llIndx = llIndx + lpvFreeClusterAddr + 1
      endif
    endwhile
    if blIsAddrOk = TRUE then
      blIsFound = TRUE
      lpvFreeClusterAddr = lpvFreeClusterAddr + llAddrCounter
      lpvFreeClusterAddr = lpvFreeClusterAddr * bFATClusterPosDiv / bFATClusterPosMult
    else
      llAddrCounter = llAddrCounter + lpvFreeClusterAddr + 1
    endif
  endwhile
end


sub lFATGetNextFreeAddrInCluster( long lpClusterAddr, lpBlockSize, lpBlockNum; &
  byte bpIsRoot, bpIsDir; var long lpvFreeAddr )
  byte blIsErased, blIsOk
  long llSectorAddr, llCurTestedSize, llCurTestedCount, llWholeTestedCount, llWantedAreaSize
  string slTestedBlock$(MAX_TEST_BLOCK_SIZE)
  string slErasedByteStr$(2)

#ifdef FS_SECURE_INIT
  slTestedBlock$ = "" '$JK
  slErasedByteStr$ = "" '$JK
#endif	'' FS_SECURE_INIT

  llWantedAreaSize = lpBlockSize * lpBlockNum
  llCurTestedCount = 0
  llWholeTestedCount = 0
  lpvFreeAddr = -1
  if bpIsRoot = TRUE then
    llSectorAddr = lpClusterAddr
    slErasedByteStr$ = HD_ERASED_BYTE_ROOT_STR
  else
    call lFATClusterToSectorAddr( lpClusterAddr, llSectorAddr )
    slErasedByteStr$ = HD_ERASED_BYTE_STR
  endif
  while llWholeTestedCount < lBytesPerCluster
    blIsErased = TRUE
    lpvFreeAddr = llSectorAddr
    llCurTestedCount = 0
    while blIsErased = TRUE and llCurTestedCount < lpBlockSize and &
      llCurTestedCount < lBytesPerCluster
        if llCurTestedCount + MAX_TEST_BLOCK_SIZE < llWantedAreaSize then
          llCurTestedSize = MAX_TEST_BLOCK_SIZE
      	else
          llCurTestedSize = llWantedAreaSize - llCurTestedCount
      	endif
      	call sHdReadString( llSectorAddr, llCurTestedSize, slTestedBlock$, blIsOk )
      	slTestedBlock$ = remove$( slTestedBlock$, slErasedByteStr$ )
      	if len( slTestedBlock$ ) = 0 or &
          ( bpIsDir = TRUE and nfroms( slTestedBlock$, 0, 1 ) = DIR_ERASED_ID ) then
            llSectorAddr = llSectorAddr + llCurTestedSize
            llCurTestedCount = llCurTestedCount + llCurTestedSize
      	else
          blIsErased = FALSE
          llSectorAddr = llSectorAddr + lpBlockSize
          llCurTestedCount = llCurTestedCount + lpBlockSize
      	endif
    endwhile
    if blIsErased = TRUE then
      llWholeTestedCount = lBytesPerCluster
    else
      llWholeTestedCount = llWholeTestedCount + llCurTestedCount
    endif
  endwhile
  if blIsErased = FALSE or llCurTestedCount <> llWantedAreaSize then
    lpvFreeAddr = -1
  endif
end


sub vFATEraseDataCluster( long lpCurDataCluster )
  long llSectorAddr

  call lFATClusterToSectorAddr( lpCurDataCluster, llSectorAddr )
  '' write the whole cluster with ff's
  sClusterBuffer$ = fill$( FAT_FILL_SECT_ERASE_PAT_STR, lBytesPerCluster )
  call vHdWriteString( llSectorAddr, 0, lBytesPerCluster, sClusterBuffer$ )
end


sub vFATEraseChainCluster( long lpCurChainCluster )
  call lFATWriteCluster( lpCurChainCluster, FAT_EMPTY_CLUSTER )
end


sub sFATConvNameTo83Form( string spSrcName$; var string spvDestName$; byte bpToUpper )
  byte blTermPosFound
  long llTermPos, llExtLen
  spvDestName$ = fill$( DIR_NAME_FILL_PAT, DIR_NAME_SIZE )
  if spSrcName$ = "." or spSrcName$ = ".." then
    spvDestName$ = stos$( spvDestName$, 0, spSrcName$, 0, len( spSrcName$ ) )
    return
  endif
  llTermPos = instr( DIR_EXT_TERM, spSrcName$, 0, 4 )
  if llTermPos <> -1 then
    blTermPosFound = TRUE
  else
    llTermPos = len( trim$( spSrcName$, " ", -1 ) )
    blTermPosFound = FALSE
  endif

  '' is the file name longer than 8 characters
  if llTermPos > DIR_EXT_POS then
    spvDestName$ = stos$( spvDestName$, 0, spSrcName$, 0, DIR_EXT_POS - 1 )
    spvDestName$ = stos$( spvDestName$, DIR_EXT_POS - 2, DIR_NAME_CONT_CHAR, 0, 1 )
    spvDestName$ = stos$( spvDestName$, DIR_EXT_POS - 1, "1", 0, 1 )
  else
    spvDestName$ = stos$( spvDestName$, 0, spSrcName$, 0, llTermPos )
  endif

  if blTermPosFound = TRUE then
    llExtLen = len( spSrcName$ ) - ( llTermPos + 1 )
    llExtLen = limit( llExtLen, 0, DIR_EXT_POS )
    spvDestName$ = stos$( spvDestName$, DIR_EXT_POS, spSrcName$, llTermPos+1, llExtLen )
  endif
  if bpToUpper = TRUE then
    spvDestName$ = upper$( spvDestName$ )
  endif
end


sub sFATConvNameToSearchForm( string spSrcName$; var string spvDestName$; byte bpToUpper )
  byte blIsTermFound
  long llTermPos, llExtLen
  spvDestName$ = ""
  if len( trim$( spSrcName$, " ", 0 ) ) = 0 then
    return
  endif
  blIsTermFound = TRUE
  llTermPos = instr( DIR_EXT_TERM, spSrcName$, 0, 4 )
  if llTermPos = -1 then
    spvDestName$ = trim$( spSrcName$, " ", -1 )
    llTermPos = len( spvDestName$ )
    blIsTermFound = FALSE
  endif
  if llTermPos > 8 then
    spvDestName$ = left$( spSrcName$, 6 )
    spvDestName$ = spvDestName$ + "*"
  else
    '' llTermPos = 0 for the search pattern ".*", convert it to "*.*"
    if llTermPos = 0 then
      spvDestName$ = spvDestName$ + "*"
    else
      spvDestName$ = left$( spSrcName$, llTermPos )
    endif
  endif
  if blIsTermFound = TRUE then
    llExtLen = len(spSrcName$) - llTermPos - 1
    if llTermPos > 8 or llExtLen > 3 then
      spvDestName$ = spvDestName$ + DIR_EXT_TERM + "*"
    else
      spvDestName$ = spvDestName$ + DIR_EXT_TERM + mid$( spSrcName$, llTermPos + 1, llExtLen )
    endif
  endif
  if spvDestName$ = "" and spSrcName$ <> "" then
    spvDestName$ = spSrcName$
  endif
  if bpToUpper = TRUE then
    spvDestName$ = upper$( spvDestName$ )
  endif
end


'' The sub finds free memory for a new dir entry, fills dir entry globals with
'' correct values and writes the new dir entry to the found location in memory
sub bFATMakeNewDirEntry( long lpParentDirAddr; string spAliasName$; string spLongName$; byte bpAttr; var byte bpvIsMade )
  byte blIsAvailable, blIsInRoot, blBreak, blIsEndCluster
  long llDirAddr, llAddrInCluster, llClusterForData, llDirCluster, &
    llAddClusterSize, llBlockNum, llNextCluster

  bpvIsMade = FALSE
  blIsInRoot = TRUE
  llDirAddr = lpParentDirAddr
  if llDirAddr <> lRootAreaStart then
    '' get dir entry of parent directory located at "lpParentDirAddr"
    call bFATGetDirEntry( llDirAddr, blIsAvailable )
    '' find next free address in the cluster that was pointed by "wDirCluster" of
    '' dir entry of the parent directory
    llDirAddr = wDirCluster
    blIsInRoot = FALSE
  endif
  '' calculate number of slots for long file name
  call lCalcNumOfSlotsLFN( spLongName$, llBlockNum )
  llBlockNum = llBlockNum + 1
  '' find next free place for dir entries
  blBreak = FALSE
  while blBreak = FALSE
    '' for root dir "llDirAddr" is address, for another dir "llDirAddr" is number of cluster
    call lFATGetNextFreeAddrInCluster( llDirAddr, DIR_ENTRY_SIZE, llBlockNum, blIsInRoot, TRUE, llAddrInCluster )
    if llAddrInCluster >= 0 then
      blBreak = TRUE
    else
      if blIsInRoot = TRUE then
        '' check, whether the place in the root directory is exhausted
        if llDirAddr + lBytesPerCluster < lRootSectors * wBpbBytesPerSector then
          blBreak = TRUE
        else
          llDirAddr = llDirAddr + lBytesPerCluster
        endif
      else
        call lFATGetNextClusterAddr( llDirAddr, 1, llNextCluster )
        call bFATIsEndCluster( llNextCluster, blIsEndCluster )
        if llNextCluster <> -1 and blIsEndCluster = FALSE then
          llDirAddr = llNextCluster
        else
          call lFATGetNextFreeCluster( llDirAddr, llNextCluster )
          if llNextCluster > 0 then
            call lFATWriteCluster( llDirAddr, llNextCluster )
            call lFATWriteCluster( llNextCluster, alFATEndClusterMaxId( bFATType ) )
            llDirAddr = llNextCluster
            call vFATEraseDataCluster( llDirAddr )
            call lFATClusterToSectorAddr( llDirAddr, llAddrInCluster )
          endif
          blBreak = TRUE
        endif
      endif
    endif
  endwhile
  '' is the next free address in cluster found?
  if llAddrInCluster >= 0 then
    call wGetCurrentDateTime( wDirCreateDate, wDirCreateTime )
    wDirAccessDate = wDirCreateDate
    wDirDate = wDirCreateDate
    wDirTime = wDirCreateTime
    '' not used now
    bDirCreate10MS = 0
    bDirReserved = 0
    lDirSize = 0
    '' DIR_ATTR_ARCHIVE flasg set only for files, not for directories
    '' DIR_ATTR_ARCHIVE means that of the file was made no regular backup
    if ( bpAttr bitand DIR_ATTR_DIRECTORY ) = 0 then
      bDirAttr = bpAttr bitor DIR_ATTR_ARCHIVE
    else
      bDirAttr = bpAttr
    endif
    if bFATType = FAT32 then
    else
      '' convert file/dir name into 8.3 form for FAT12, FAT16
      call sFATConvNameTo83Form( spLongName$, sDirName$, TRUE )
      wDirClusterHigh = 0
      '' the beginning cluster is always defined for a new directory,
      '' because the entries "." and ".." must be made
      if ( bpAttr bitand DIR_ATTR_DIRECTORY ) <> 0 then
        '' reserve a new cluster (in FAT table)
    	call lFATGetNextFreeCluster( 0, llDirCluster )
        if llDirCluster = -1 then
          '' error: there is no more free cluster
          return
        else
          wDirCluster = llDirCluster
          call lFATWriteCluster( llDirCluster, alFATEndClusterMaxId( bFATType ) )
        endif
      else
        wDirCluster = 0
      endif
    endif
    '' write the long name slots
    call vFATWriteSlotsLFN( spLongName$, sDirName$, llAddrInCluster )
    '' write the new dir entry
    call bFATWriteDirEntry( llAddrInCluster )

    '' make the entries "." and ".." for directory
    if ( bpAttr bitand DIR_ATTR_DIRECTORY ) <> 0 then
''''      call lFATGetNextFreeAddrInCluster( llDirCluster, DIR_ENTRY_SIZE, 1, blIsInRoot, llAddrInCluster )
      call lFATClusterToSectorAddr( llDirCluster, llAddrInCluster )
      wDirCluster = llDirCluster

      '' "." entry points to itself
      '' convert file/dir name into 8.3 form for FAT12, FAT16
      call sFATConvNameTo83Form( ".", sDirName$, FALSE )
      '' write the new dir entry
      call bFATWriteDirEntry( llAddrInCluster )

      '' ".." entry points to the parent directory
      '' convert file/dir name into 8.3 form for FAT12, FAT16
      call sFATConvNameTo83Form( "..", sDirName$, FALSE )
      wDirCluster = llDirAddr
      '' write the new dir entry
      llAddrInCluster = llAddrInCluster + DIR_ENTRY_SIZE
      call bFATWriteDirEntry( llAddrInCluster )

      '' fill the rest of the cluster where "." and ".." are located with FAT_FILL_SECT_PAT_STR
      llAddClusterSize = lBytesPerCluster - DIR_ENTRY_SIZE * 2
      sClusterBuffer$ = fill$( FAT_FILL_SECT_PAT_STR, llAddClusterSize )
      set_len$( sClusterBuffer$, llAddClusterSize )
      llAddrInCluster = llAddrInCluster + DIR_ENTRY_SIZE
      call vHdWriteString( llAddrInCluster, 0, llAddClusterSize, sClusterBuffer$ )
    endif
    bpvIsMade = TRUE
  endif
end


sub bFATIsPathAbsolute( string spPathName$; var byte bpvIsPathAbs )
  string slFirstChar$(1)

#ifdef FS_SECURE_INIT
  slFirstChar$ = "" '$JK
#endif	'' FS_SECURE_INIT

  bpvIsPathAbs = FALSE
  if spPathName$ = ROOT_DIR then
    bpvIsPathAbs = TRUE
  else
    slFirstChar$ = left$( spPathName$, 1 )
    if slFirstChar$ = PATH_LIM_DOS or slFirstChar$ = PATH_LIM_UNIX then
      bpvIsPathAbs = TRUE
    endif
  endif
end


'' The sub returns the part of full file name included between two terminators
sub sFATGetNextPathName( string spFullFileName$; var long lpvStartPos; &
  var string spvNextPath$; byte bpPathType )
  byte blIsFound
  long llLimPos, llNameLen
  string sl83Name$(DIR_NAME_FULL_SIZE)

#ifdef FS_SECURE_INIT
  sl83Name$ = "" '$JK
#endif	'' FS_SECURE_INIT

  blIsFound = FALSE
  llNameLen = len( spFullFileName$ )
  while blIsFound = FALSE
    llLimPos = instr( PATH_LIM_DOS, spFullFileName$, lpvStartPos, 4 )
    if llLimPos = -1 then
      llLimPos = instr( PATH_LIM_UNIX, spFullFileName$, lpvStartPos, 4 )
    endif
    if llLimPos = lpvStartPos then
      lpvStartPos = lpvStartPos + 1
    else
      blIsFound = TRUE
    endif
  endwhile
  if llLimPos <> -1 then
    spvNextPath$ = mid$( spFullFileName$, lpvStartPos, llLimPos - lpvStartPos )
    lpvStartPos = llLimPos + 1
  else
    spvNextPath$ = mid$( spFullFileName$, lpvStartPos, llNameLen - lpvStartPos )
    lpvStartPos = len( spFullFileName$ )
  endif
  if bpPathType <> PATH_NAME_TYPE_LONG and ( bFATType = FAT12 or bFATType = FAT16 ) then
    switch bpPathType
      case PATH_NAME_TYPE_SHORT:
        call sFATConvNameTo83Form( spvNextPath$, sl83Name$, TRUE )
      case PATH_NAME_TYPE_FOR_SEARCH:
        call sFATConvNameToSearchForm( spvNextPath$, sl83Name$, TRUE )
      default:
        sl83Name$ = spvNextPath$
    endswitch
    spvNextPath$ = sl83Name$
  endif
end


'' The sub searches for address of file info fields describing the file referenced by
'' "spFullFileName$"; the search begins from "lpStartAddr";
'' the found file info fields will be copied to some global vars in the sub
'' "lFATFindDirEntryInCluster" called here
sub lFATFindFileInfo( string spFullFileName$; long lpStartAddr; var long lpvFileInfoAddr )
  string slName$( MAX_LONGNAME_SIZE ), slFullName$( MAX_LONGNAME_SIZE )
  long llCurPos, llNameLen, llCurAddr, llCurCluster, llNextCluster
  byte blInRootArea, blIsValid, blIsEnd, blIsPathAbs

#ifdef FS_SECURE_INIT
  slName$ = "" '$JK
  slFullName$ = "" '$JK
#endif	'' FS_SECURE_INIT

  llCurPos = 0
  blInRootArea = TRUE
  slFullName$ = spFullFileName$
  call bFATIsPathAbsolute( slFullName$, blIsPathAbs )
  if blIsPathAbs = TRUE then
    '' the search begins from the root dir
    llCurAddr = lRootAreaStart
  else
    if lpStartAddr = 0 then
      '' the search begins from the root dir
      llCurAddr = lRootAreaStart
      blInRootArea = TRUE
      if sCurrentDir$ <> ROOT_DIR then
        '' make full path
        slFullName$ = trim$( sCurrentDir$, " ", 0 ) + PATH_LIM_DOS + spFullFileName$
      endif
    else
      llCurAddr = lpStartAddr
      if lCurrentDirFATAddr <> lRootAreaStart then
        blInRootArea = FALSE
      endif
    endif
  endif
  llNameLen = len( slFullName$ )

  llCurCluster = 0
  '' loop to search for the file through the full path
  while llCurPos < llNameLen
    '' get the next part of path
''''    call sFATGetNextPathName( slFullName$, llCurPos, slName$, PATH_NAME_TYPE_SHORT )
    call sFATGetNextPathName( slFullName$, llCurPos, slName$, PATH_NAME_TYPE_FOR_SEARCH )
    lpvFileInfoAddr = -2
    '' -2 is returned, if the path was not found and the end of cluster was reached
    while lpvFileInfoAddr = -2
      call lFATFindDirEntryInCluster( slName$, llCurAddr, lpvFileInfoAddr )
      '' if the path was found, read the cluster for the child subdirectory or for the file contents
      if lpvFileInfoAddr >= 0 then
        llCurCluster = wDirCluster
	call lFATClusterToSectorAddr( llCurCluster, llCurAddr )
        blInRootArea = FALSE
      else
        '' path was not yet found; try to search for in the next cluster
        '' ??? Should we search for DirEntry in the next linked (FAT) cluster or not ?
        if lpvFileInfoAddr = -2 then
          '' if we are in root area, don't reference to FAT in looking for next cluster;
          '' size of root area is fixed and all root area clusters lay consecutively
          if blInRootArea = TRUE then
            if llCurCluster < lRootSectors / bBpbSectorsPerCluster - 1 then
              llCurCluster = llCurCluster + 1
              llCurAddr = llCurCluster * lBytesPerCluster + lRootAreaStart
            else
              lpvFileInfoAddr = -1
            endif
          else
	    '' if we are not in the root area, look for next cluster in FAT
            call lFATGetNextClusterAddr( llCurCluster, 1, llNextCluster )
            llCurCluster = llNextCluster
            call bFATIsEndCluster( llCurCluster, blIsEnd )
            if llCurCluster <> -1 and blIsEnd = FALSE then
              call lFATClusterToSectorAddr( llCurCluster, llCurAddr )
            else
              lpvFileInfoAddr = -1
            endif
          endif
        endif
      endif
    endwhile
    '' if the name was not found, break the loop
    if lpvFileInfoAddr = -1 then
      llCurPos = llNameLen
    endif
  endwhile
end


'' Look up the cluster chain to find to which cluster "lpAddrInFile" belongs
sub lFATFindSectorByAddr( long lpAddrInFile; var long lpvCurCluster, lpvCurAddr )
  byte blIsAvailable
  long llAddrCounter, llNextCluster
  llAddrCounter = 0
  lpvCurAddr = -1
  blIsAvailable = FALSE
  while blIsAvailable = FALSE and llAddrCounter <= lpAddrInFile
    if llAddrCounter <= lpAddrInFile and lpAddrInFile < llAddrCounter + lBytesPerCluster then
      call lFATClusterToSectorAddr( lpvCurCluster, lpvCurAddr )
      blIsAvailable = TRUE
    else
      call lFATGetNextClusterAddr( lpvCurCluster, 1, llNextCluster )
      lpvCurCluster = llNextCluster
      llAddrCounter = llAddrCounter + lBytesPerCluster
    endif
  endwhile
end


sub bFATReadFile( long lpFileFatEntryPos; var long lpvPosInFile; var string spvBuffer$; long lpSize; &
  var byte bpvIsRead )
  byte blIsDone, blIsEndCluster, blIsReadOk
  word wlDate, wlTime
  long llCurClusterAddr, llNextClusterAddr, llIndx, llRead, llSectorAddr, &
    llCurPos, llCurSizeToRead, llCurSizeToWriteCorrected, llPrevFileSize

  bpvIsRead = FALSE
  spvBuffer$ = ""
  call bFATGetDirEntry( lpFileFatEntryPos, blIsDone )
  if blIsDone = TRUE then
    llPrevFileSize = lDirSize
    '' calculate to which cluster "lpvPosInFile" belongs
    llIndx = lpvPosInFile / lBytesPerCluster
''    if llIndx > 0 and mod( lpvPosInFile, lBytesPerCluster ) = 0 then
''      llIndx = llIndx - 1
''    endif
    llCurClusterAddr = wDirCluster
    if wDirCluster > 0 then
#ifndef FAT_IN_RAM
      loop llIndx
        call lFATGetNextClusterAddr( llCurClusterAddr, 1, llNextClusterAddr )
        llCurClusterAddr = llNextClusterAddr
      endloop
#else
      call lFATGetNextClusterAddr( llCurClusterAddr, llIndx, llNextClusterAddr )
      llCurClusterAddr = llNextClusterAddr
#endif
    endif
    '' calculate the current position inside of cluster
    llCurPos = mod( lpvPosInFile, lBytesPerCluster )
    '' calculate the size of the area remained inside of cluster
    llCurSizeToRead = lBytesPerCluster - llCurPos

    '' stay in the loop until all data were written
    llRead = 0
    while llRead < lpSize
      if lpvPosInFile + llRead >= llPrevFileSize then
        '' error: attempt to read from the position over the file end
        bpvIsRead = FALSE
        return
      else
        if llRead > 0 and mod( lpvPosInFile + llRead, lBytesPerCluster ) = 0 then
	  '' get the next cluster addr from the cluster chain
          call lFATGetNextClusterAddr( llCurClusterAddr, 1, llNextClusterAddr )
          llCurClusterAddr = llNextClusterAddr
          if llCurClusterAddr = -1 then
            '' error: cluster chain doesn't correspond with the file size
          endif
        endif
      endif

      '' read the next possibly long part of data from the found cluster
      call lFATClusterToSectorAddr( llCurClusterAddr, llSectorAddr )
      llCurPos = llCurPos + llSectorAddr

      '' calculate the size of data to write
      if lpSize - llRead > lBytesPerCluster then
        llCurSizeToRead = lBytesPerCluster
      else
        llCurSizeToRead = lpSize - llRead
      endif

      call sHdReadString( llCurPos, llCurSizeToRead, sClusterBuffer$, blIsReadOk )
      if blIsReadOk = TRUE then
        spvBuffer$ = stos$( spvBuffer$, llRead, sClusterBuffer$, llCurSizeToRead )
      endif
      llRead = llRead + llCurSizeToRead
      llCurPos = mod( lpvPosInFile + llRead, lBytesPerCluster )
    endwhile
    lpvPosInFile = lpvPosInFile + llRead
    lDirSize = llPrevFileSize
    call wGetCurrentDateTime( wlDate, wlTime )
    wDirAccessDate = wlDate
    call bFATWriteDirEntry( lpFileFatEntryPos )
    if llRead > 0 then
      set_len$( spvBuffer$, llRead )
      bpvIsRead = TRUE
    endif
  endif
end


sub bFATWriteFile( long lpFileFatEntryPos; var long lpvPosInFile; var long lpvLastCluster; &
  string spBuffer$; long lpSize; var byte bpvIsWritten )
  byte blIsDone, blIsEndCluster
  word wlDate, wlTime
  long llCurClusterAddr, llNextClusterAddr, llIndx, llWritten, llSectorAddr, &
    llCurPos, llCurSizeToWrite, llCurSizeToWriteCorrected, &
    llPrevClusterAddr, llPrevFileSize, llRemainedInCluster

  bpvIsWritten = FALSE
  call bFATGetDirEntry( lpFileFatEntryPos, blIsDone )
  if blIsDone = TRUE then
    '' check read-only attribute of the file
    if ( bDirAttr bitand DIR_ATTR_READONLY ) = DIR_ATTR_READONLY then
      bpvIsWritten = FALSE
      return
    endif
    llPrevFileSize = lDirSize
    '' calculate to which cluster "lpvPosInFile" belongs
    llIndx = lpvPosInFile / lBytesPerCluster
    if llIndx > 0 and mod( lpvPosInFile, lBytesPerCluster ) = 0 and lpvPosInFile >= llPrevFileSize then
      llIndx = llIndx - 1
    endif
    llCurClusterAddr = wDirCluster
    if wDirCluster > 0 then
#ifndef FAT_IN_RAM
      loop llIndx
        call lFATGetNextClusterAddr( llCurClusterAddr, 1, llNextClusterAddr )
        llCurClusterAddr = llNextClusterAddr
      endloop
#else
      if llIndx > 0 then
        if lpvLastCluster = 0 then
          lpvLastCluster = wDirCluster
          llCurClusterAddr = lpvLastCluster
          call lFATGetNextClusterAddr( llCurClusterAddr, llIndx, llNextClusterAddr )
          llCurClusterAddr = llNextClusterAddr
        else
          llCurClusterAddr = lpvLastCluster
        endif
      endif
#endif
    else
      if lpvPosInFile <> 0 then
        '' intern error: incorrect positioning inside of file
        return
      endif
    endif
    llPrevClusterAddr = llCurClusterAddr
    '' calculate the current position inside of cluster
    llCurPos = mod( lpvPosInFile, lBytesPerCluster )
    '' calculate the size of the area remained inside of cluster
    llRemainedInCluster = lBytesPerCluster - llCurPos

    '' stay in the loop until all data were written
    llWritten = 0
    while llWritten < lpSize
      if lpvPosInFile + llWritten >= llPrevFileSize and &
        mod( lpvPosInFile + llWritten, lBytesPerCluster ) = 0 then
        '' reserve a new cluster (in FAT table)
        call lFATGetNextFreeCluster( llCurClusterAddr, llNextClusterAddr )
        llCurClusterAddr = llNextClusterAddr
        if llCurClusterAddr <> -1 then
          call lFATWriteCluster( llCurClusterAddr, alFATEndClusterMaxId( bFATType ) )
        else
          '' error: there is no more free cluster
        endif
        if llPrevClusterAddr > 0 then
          '' write the address of the new cluster into the cluster chain
          call lFATWriteCluster( llPrevClusterAddr, llCurClusterAddr )
        endif
        llPrevClusterAddr = llCurClusterAddr
        llRemainedInCluster = lBytesPerCluster
      else
        if llWritten > 0 and mod( lpvPosInFile + llWritten, lBytesPerCluster ) = 0 then
	  '' get the next cluster addr from the cluster chain
          call lFATGetNextClusterAddr( llCurClusterAddr, 1, llNextClusterAddr )
          llCurClusterAddr = llNextClusterAddr
          if llCurClusterAddr = -1 then
            '' error: cluster chain doesn't correspond with the file size
          endif
          llRemainedInCluster = lBytesPerCluster
        endif
        llPrevClusterAddr = llCurClusterAddr
      endif

      '' set new start cluster for the file, if written at first time
      if wDirCluster = 0 then
        wDirCluster = llCurClusterAddr
      endif

      '' write the next possibly long part of data to the found/created cluster
      call lFATClusterToSectorAddr( llCurClusterAddr, llSectorAddr )
      llCurPos = llCurPos + llSectorAddr

      '' calculate the size of data to write
      if lpSize - llWritten > lBytesPerCluster then
        llCurSizeToWrite = lBytesPerCluster
      else
        llCurSizeToWrite = lpSize - llWritten
      endif
      if llCurSizeToWrite > llRemainedInCluster then
        llCurSizeToWrite = llRemainedInCluster
      endif
      '' fill the unused part of the sector with fill pattern
      llCurSizeToWriteCorrected = llCurSizeToWrite
      if lpvPosInFile + llWritten  + llCurSizeToWrite > lDirSize and &
        mod( llCurSizeToWrite, wBpbBytesPerSector ) <> 0 then
          llCurSizeToWriteCorrected = llCurSizeToWrite + &
            ( wBpbBytesPerSector - mod( llCurSizeToWrite, wBpbBytesPerSector ) )
          sClusterBuffer$ = fill$( FAT_FILL_SECT_PAT_STR, llCurSizeToWriteCorrected )
      endif

      sClusterBuffer$ = stos$( sClusterBuffer$, 0, spBuffer$, llWritten, llCurSizeToWrite )
      set_len$( sClusterBuffer$, llCurSizeToWriteCorrected )
      call vHdWriteString( llCurPos, 0, llCurSizeToWriteCorrected, sClusterBuffer$ )
      llWritten = llWritten + llCurSizeToWrite
      llRemainedInCluster = llRemainedInCluster - llCurSizeToWrite
      llCurPos = mod( lpvPosInFile + llWritten, lBytesPerCluster )
    endwhile
    lpvPosInFile = lpvPosInFile + llWritten
    '' update the size of the file and the date/time in the directory entry		
    if lpvPosInFile > llPrevFileSize then
      lDirSize = lpvPosInFile
    else
      lDirSize = llPrevFileSize
    endif
    call wGetCurrentDateTime( wlDate, wlTime )
    wDirAccessDate = wlDate
    wDirDate = wlDate
    wDirTime = wlTime
    call bFATWriteDirEntry( lpFileFatEntryPos )
    if llWritten > 0 then
      bpvIsWritten = TRUE
    endif
    lpvLastCluster = llCurClusterAddr
  endif
end


sub lFATGetNewFilePtr( long lpFileFatEntryPos, lpCurFilePtr, lpOffset; byte bpWhence;&
  var long lpvNewFilePtr )
  byte blIsDone
  lpvNewFilePtr = -1
  call bFATGetDirEntry( lpFileFatEntryPos, blIsDone )
  if blIsDone = TRUE then
    switch bpWhence
      case FILE_BEGIN:
        if lpOffset >= 0 and lpOffset <= lDirSize then
    	  lpvNewFilePtr = lpOffset
        endif
      case FILE_CURRENT:
        if lpCurFilePtr + lpOffset <= lDirSize and lpCurFilePtr + lpOffset >= 0 then
    	  lpvNewFilePtr = lpCurFilePtr + lpOffset
        endif
      case FILE_END:
        if lpOffset <= 0 and lDirSize + lpOffset >= 0 then
    	  lpvNewFilePtr = lDirSize + lpOffset
        endif
    endswitch
  endif
end


sub lFATGetDiskFreeSpace( var long lpvFreeSpace )
#ifdef FAT_IN_RAM
  long llUsedClusters
  call lFATTableGetUsedClusters( llUsedClusters )
  lpvFreeSpace = lDataSectors * wBpbBytesPerSector - llUsedClusters * lBytesPerCluster
#endif
end


sub bFATSetup( var byte bpvIsFATSetupOk )
  byte blFATBootSectorFound, blFATTableOk
  long llFATBootSectAddrPhys, llFATBootSectAddrLog

  blFATTableOk = TRUE
  bpvIsFATSetupOk = FALSE
  llFATBootSectAddrLog = 0h
#ifdef FS_SECURE_INIT
  call vFATInitGlobalVars()
#endif	'' FS_SECURE_INIT
  call lFATGetBootSectAddr( llFATBootSectAddrLog, blFATBootSectorFound )

  if blFATBootSectorFound = TRUE then
    call vFATGetBootRecord( llFATBootSectAddrLog )
#ifdef DIR_ENTRIES_IN_RAM
    call vFATInitDirEntriesInRam()
#endif
#ifdef FAT_IN_RAM
    call bFATInitTable( blFATTableOk )
#endif
    if blFATTableOk = TRUE then
      bpvIsFATSetupOk = TRUE
    endif
  endif
end


#ifdef FS_SECURE_INIT
sub vFATInitGlobalVars()
  word wlIndx, wlSecIndx

  '' Master Boot for FAT12 and for FAT16 volumes
  bMbx86boot = FS_NUM_VARS_INIT_VALUE     	'' x86 defaut boot partition
  bMbStartHead = FS_NUM_VARS_INIT_VALUE		'' start head
  bMbStartSector = FS_NUM_VARS_INIT_VALUE	'' start sector
  bMbStartCylinder = FS_NUM_VARS_INIT_VALUE	'' start cylinder
  bMbPartitionType = FS_NUM_VARS_INIT_VALUE	'' partition type
  bMbEndHead = FS_NUM_VARS_INIT_VALUE       	'' end head
  bMbEndSector = FS_NUM_VARS_INIT_VALUE		'' end sector
  bMbEndCylinder = FS_NUM_VARS_INIT_VALUE	'' end cylinder
  lMbLogStartSector = FS_NUM_VARS_INIT_VALUE	'' logical start sector
  lMbNumSectors = FS_NUM_VARS_INIT_VALUE	'' number of all sectors
  wMbSignature = FS_NUM_VARS_INIT_VALUE		'' signature word

  '' BPB for FAT12 and for FAT16 volumes
  wBpbBytesPerSector = FS_NUM_VARS_INIT_VALUE
  bBpbSectorsPerCluster = FS_NUM_VARS_INIT_VALUE
  wBpbReservedSectors = FS_NUM_VARS_INIT_VALUE
  bBpbNumberOfFATs = FS_NUM_VARS_INIT_VALUE
  wBpbRootEntries = FS_NUM_VARS_INIT_VALUE
  lBpbTotalSectors = FS_NUM_VARS_INIT_VALUE
  bBpbMedia = FS_NUM_VARS_INIT_VALUE
  wBpbSectorsPerFAT = FS_NUM_VARS_INIT_VALUE
  wBpbSectorsPerTrack = FS_NUM_VARS_INIT_VALUE
  wBpbHeadsPerCylinder = FS_NUM_VARS_INIT_VALUE
  lBpbHiddenSectors = FS_NUM_VARS_INIT_VALUE
  lBpbTotalSectorsBig = FS_NUM_VARS_INIT_VALUE

  '' BPB for FAT32 volumes (additional fields)
  lBpb32SectorsPerFAT = FS_NUM_VARS_INIT_VALUE
  wBpb32Flags = FS_NUM_VARS_INIT_VALUE
  wBpb32Version = FS_NUM_VARS_INIT_VALUE
  lBpb32RootCluster = FS_NUM_VARS_INIT_VALUE
  wBpb32InfoSector = FS_NUM_VARS_INIT_VALUE
  wBpb32BootBackupStart = FS_NUM_VARS_INIT_VALUE
  abBpb32Reserved$ = FS_STR_VARS_INIT_VALUE

  '' Boot Sector layout for FAT12, FAT16, FAT32
  sBsJmp$ = FS_STR_VARS_INIT_VALUE
  sBsOemName$ = FS_STR_VARS_INIT_VALUE
  bBsDriveNumber = FS_NUM_VARS_INIT_VALUE
  bBsUnused = FS_NUM_VARS_INIT_VALUE
  bBsExtBootSignature = FS_NUM_VARS_INIT_VALUE
  lBsSerialNumber = FS_NUM_VARS_INIT_VALUE
  sBsVolumeLabel$ = FS_STR_VARS_INIT_VALUE
  sBsFileSystem$ = FS_STR_VARS_INIT_VALUE
  sBsBootCode$ = FS_STR_VARS_INIT_VALUE

  '' FAT32 File System Information Structure
  lFsInfoSignature = FS_NUM_VARS_INIT_VALUE
  lFsInfoFreeClusterCount = FS_NUM_VARS_INIT_VALUE
  lFsInfoNextCluster = FS_NUM_VARS_INIT_VALUE
  sFsInfoReserved$ = FS_STR_VARS_INIT_VALUE

  '' Directory Entry format
  sDirName$ = FS_STR_VARS_INIT_VALUE
  bDirAttr = FS_NUM_VARS_INIT_VALUE
  bDirReserved = FS_NUM_VARS_INIT_VALUE
  bDirCreate10MS = FS_NUM_VARS_INIT_VALUE
  wDirCreateTime = FS_NUM_VARS_INIT_VALUE
  wDirCreateDate = FS_NUM_VARS_INIT_VALUE
  wDirAccessDate = FS_NUM_VARS_INIT_VALUE
  wDirClusterHigh = FS_NUM_VARS_INIT_VALUE
  wDirTime = FS_NUM_VARS_INIT_VALUE
  wDirDate = FS_NUM_VARS_INIT_VALUE
  wDirCluster = FS_NUM_VARS_INIT_VALUE
  lDirSize = FS_NUM_VARS_INIT_VALUE

  '' LFN (Long File Name) Entry format
  bLfnSequence = FS_NUM_VARS_INIT_VALUE
  sLfnName1$ = FS_STR_VARS_INIT_VALUE
  bLfnAttr = FS_NUM_VARS_INIT_VALUE
  bLfnReserved = FS_NUM_VARS_INIT_VALUE
  bLfnChecksum = FS_NUM_VARS_INIT_VALUE
  sLfnName2$ = FS_STR_VARS_INIT_VALUE
  wLfnCluster = FS_NUM_VARS_INIT_VALUE
  sLfnName3$ = FS_STR_VARS_INIT_VALUE

  lFATSectors = FS_NUM_VARS_INIT_VALUE
  lDataSectors = FS_NUM_VARS_INIT_VALUE
  lRootSectors = FS_NUM_VARS_INIT_VALUE
  lTotalClusters = FS_NUM_VARS_INIT_VALUE
  lBytesPerCluster = FS_NUM_VARS_INIT_VALUE
  lBytesPerFAT = FS_NUM_VARS_INIT_VALUE
  lFATStart = FS_NUM_VARS_INIT_VALUE
  lRootAreaStart = FS_NUM_VARS_INIT_VALUE
  bFATType = FS_NUM_VARS_INIT_VALUE
  '' Two variables to avoid operations with reals,
  '' while calculating the position of a cluster inside of FAT
  bFATClusterPosMult = FS_NUM_VARS_INIT_VALUE
  bFATClusterPosDiv = FS_NUM_VARS_INIT_VALUE

'' Buffer to temporary save data of a cluster
  sClusterBuffer$ = FS_STR_VARS_INIT_VALUE

  for wlIndx = 0 to TOTAL_FATS - 1
    alFATEndClusterMinId(wlIndx) = FS_NUM_VARS_INIT_VALUE
  next
  for wlIndx = 0 to TOTAL_FATS - 1
    alFATEndClusterMaxId(wlIndx) = FS_NUM_VARS_INIT_VALUE
  next
  for wlIndx = 0 to TOTAL_FATS - 1
    alFATInvalidClusterId(wlIndx) = FS_NUM_VARS_INIT_VALUE
  next

  '' address of current dir in the FAT
  sCurrentDir$ = FS_STR_VARS_INIT_VALUE
  lCurrentDirFATAddr = FS_NUM_VARS_INIT_VALUE

  '' if this flag set to TRUE, use windows-like formatting parameters
  bFATFmtWinFlag = FS_NUM_VARS_INIT_VALUE

#ifdef FAT_IN_RAM
  '' buffer to temporary keep the file allocation table
  for wlIndx = 0 to FAT_RAM_BUF_NUM - 1
    asFATRamBuffer$(wlIndx) = FS_STR_VARS_INIT_VALUE
  next
  sFATTempBuffer$ = FS_STR_VARS_INIT_VALUE
  bFATTableModified = FS_NUM_VARS_INIT_VALUE
  for wlIndx = 0 to FAT_MODIFIED_AREAS_MAX_NUM - 1
    abFATTableModifiedAreas(wlIndx) = FS_NUM_VARS_INIT_VALUE
  next
  lFATModifiedAreasNum = FS_NUM_VARS_INIT_VALUE
  lFATModifiedAreaSize = FS_NUM_VARS_INIT_VALUE
#endif

#ifdef DIR_ENTRIES_IN_RAM
  for wlIndx = 0 to DIR_ENTRIES_IN_RAM_NUM - 1
    asFATDirEntryRamBuffer$(wlIndx) = FS_STR_VARS_INIT_VALUE
  next
  bFATDirEntryRamBufModified = FS_NUM_VARS_INIT_VALUE
  lFATDirEntryRamBufSize = FS_NUM_VARS_INIT_VALUE
#endif
end
#endif	''FS_SECURE_INIT


#endif

