''**************************************************************************
'' Definition of Hardware Abstraction Layer
'' HAL is used to simplify the adaptation of the file system subroutines
'' to the working with other storage devices
''**************************************************************************

#ifndef FS_HAL_D_INC
#define FS_HAL_D_INC

'' Smart Media
'' signature: sub lHdGetMediaSize( var long lpvMediaSize )
#define lHdGetMediaSize		lSMGetMediaSize

'' signature: sub lHdReadNum( long lpAddr; byte bpSize; var long lpvNum; var byte bpvIsOk )
#define lHdReadNum		lSMReadNum

'' signature: sub lHdReadNumPhys( long lpPhysAddr; byte bpSize; var long lpvNum )
#define lHdReadNumPhys		lSMReadNumPhys

'' signature: sub sHdReadString( long lpAddr; long lpSize; var string spvBuffer$; var byte bpvIsOk )
#define sHdReadString		sSMReadString

'' signature: sub sSMReadStringPhys( long lpPhysAddr, lpLogAddr; long lpSize; var string spvBuffer$ )
#define sHdReadStringPhys	sSMReadStringPhys

'' signature: sub vHdWriteString( long lpAddr, lpPos, lpSize; var string spvBuffer$ )
#define vHdWriteString		vSMWriteString

'' signature: sub bHdInit( var byte bpvRet )
#define bHdInit			bSMInit

'' signature: sub lSMFindStringInMemBlock( string spSearchedString$; long lpStartAddr, &
''  lpMemBlockSize, lpStepSize; var long lpvStringAddr )
#define lHdFindStringInMemBlock	lSMFindStringInMemBlock

'' signature: sub bSMGetInfoSet( var string spvInfoSet$; var byte bpvIsRead )
#define bHdGetInfoSet		bSMGetInfoSet

#define lHdBytesInBlock		lSmBytesInBlock

#define lHdBytesInPage		lSmBytesInPage

#define HD_SECT_SIZE		SM_MAX_PAGE_SIZE
'' bytes used to fill erased sectors
#define HD_ERASED_BYTE_ROOT_STR	SM_ERASED_BYTE_ROOT_STR
#define HD_ERASED_BYTE_STR	SM_ERASED_BYTE_STR


'' Offsets and sizes of the fields of the hardware info set
#define HDI_MAKER_CODE_POS	SM_MAKER_CODE_POS	' <1> Maker-Code = Manufacturer
#define HDI_ID_CODE_POS		SM_ID_CODE_POS		' <1> ID-Code = Card-Id (Memory Size)
#define HDI_BYTES_IN_SPARE_POS	SM_BYTES_IN_SPARE_POS	' <2> Nr. of Bytes in Spare-Field (normal: 16 Bytes)
#define HDI_BYTES_IN_PAGE_POS	SM_BYTES_IN_PAGE_POS	' <4> Nr. of DATA Bytes in a Page (without spare-field)
#define HDI_PAGES_IN_BLOCK_POS	SM_PAGES_IN_BLOCK_POS	' <2> Nr. of Pages in a Block
#define HDI_BYTES_IN_BLOCK_POS	SM_BYTES_IN_BLOCK_POS	' <4> Nr. of DATA-Bytes in a Block
#define HDI_NO_OF_BLOCKS_POS	SM_NO_OF_BLOCKS_POS	' <2> Total Number of "Blocks" in this SmartMedia Card
#define HDI_ADR_HIGH_BLOCK_POS	SM_ADR_HIGH_BLOCK_POS	' <4> Base ADR of highest Block
#define HDI_ADR_END_POS		SM_ADR_END_POS		' <4> End-ADR = 1.ADR after the last Byte !!

#define HDI_MAKER_CODE_SIZE	SM_MAKER_CODE_SIZE	' <1> Maker-Code = Manufacturer
#define HDI_ID_CODE_SIZE   	SM_ID_CODE_SIZE		' <1> ID-Code = Card-Id (Memory Size)
#define HDI_BYTES_IN_SPARE_SIZE	SM_BYTES_IN_SPARE_SIZE	' <2> Nr. of Bytes in Spare-Field (normal: 16 Bytes)
#define HDI_BYTES_IN_PAGE_SIZE	SM_BYTES_IN_PAGE_SIZE	' <4> Nr. of DATA Bytes in a Page (without spare-field)
#define HDI_PAGES_IN_BLOCK_SIZE	SM_PAGES_IN_BLOCK_SIZE	' <2> Nr. of Pages in a Block
#define HDI_BYTES_IN_BLOCK_SIZE	SM_BYTES_IN_BLOCK_SIZE	' <4> Nr. of DATA-Bytes in a Block
#define HDI_NO_OF_BLOCKS_SIZE	SM_NO_OF_BLOCKS_SIZE	' <2> Total Number of "Blocks" in this SmartMedia Card
#define HDI_ADR_HIGH_BLOCK_SIZE	SM_ADR_HIGH_BLOCK_SIZE	' <4> Base ADR of highest Block
#define HDI_ADR_END_SIZE   	SM_ADR_END_SIZE		' <4> End-ADR = 1.ADR after the last Byte !!

'' Size of the hardware info set
#define HDI_BLOCK_SIZE		SMCARD_INFO_BLOCK_SIZE

#endif

