''**************************************************************************
'' Definitions relevant for FS API
''**************************************************************************

#ifndef FS_INX_D_INC
#define FS_INX_D_INC

'' Flag values for open file
'' file access modes
#define O_RDONLY	0001h
#define O_WRONLY	0002h
#define O_RDWR		0004h
#define O_FAM_MASK	000fh

#define O_CREAT     0100h	'' create file if the file doesn't exist and open it
'' #define O_TRUNC     0200h  	'' open with truncation
#define O_EXIST     0400h  	'' check availability of file, don't open it
#define O_APPEND    0800h  	'' file may be written to only at its end

'' #define O_TEXT      4000h	'' CR-LF translation
'' #define O_BINARY    8000h  	'' no translation

'' constants for "SetFilePointer"
#define FILE_BEGIN           0
#define FILE_CURRENT         1
#define FILE_END             2


'' File index table
''=================
'' Offsets of the different file index fields
#define FI_FLAGS_INX		0
#define FI_FAT_ENTRY_POS_INX	4
#define FI_FILE_POINTER_INX	8
#define FI_BUF_POOL_POS_INX	12
#define FI_BUF_FILE_START_INX	16
#define FI_BUF_FILE_LEN_INX	20
#define FI_LAST_CLUSTER_INX	24

'' Sizes of the different file index fields
#define FI_FLAGS_SIZE		4
#define FI_FAT_ENTRY_POS_SIZE	4
#define FI_FILE_POINTER_SIZE	4
#define FI_BUF_POOL_POS_SIZE	4
#define FI_BUF_FILE_START_SIZE	4
#define FI_BUF_FILE_LEN_SIZE	4
#define FI_LAST_CLUSTER_SIZE	4


'' The information about found file returned by "bFindFirstFile" and by "bFindNextFile"
'' is saved in a string. There are the offsets and the sizes of the data fields of
'' this string here
#define FFD_ATTR_OFFS		0
#define FFD_CREATE_TIME_MS_OFFS	1
#define FFD_CREATE_TIME_OFFS	2
#define FFD_CREATE_DATE_OFFS	4
#define FFD_ACCESS_DATE_OFFS	6
#define FFD_SIZE_OFFS		8
#define FFD_RESERVED_1_OFFS	12
#define FFD_RESERVED_2_OFFS	16
#define FFD_NAME_OFFS		20
#define FFD_EXT_OFFS		28
#define FFD_LONG_NAME_OFFS	31
#define FFD_ABRIDGED_NAME_OFFS	31

#define FFD_ATTR_SIZE		1
#define FFD_CREATE_TIME_MS_SIZE	1
#define FFD_CREATE_TIME_SIZE	2
#define FFD_CREATE_DATE_SIZE	2
#define FFD_ACCESS_DATE_SIZE	2
#define FFD_SIZE_SIZE		4
#define FFD_RESERVED_1_SIZE	4
#define FFD_RESERVED_2_SIZE	4
#define FFD_NAME_SIZE		8
#define FFD_EXT_SIZE		3
#define FFD_LONG_NAME_SIZE	MAX_LONGNAME_SIZE
#define FFD_ABRIDGED_NAME_SIZE	12

'' Size of a file find data structure (saved normally in a string) without a field
'' for the long file name
#define	FFD_STRUCT_SHORT_SIZE   32

'' Size of a file find data structure (saved normally in a string) with a shortened field
'' for the long file name (if the file name is of the "8.3" form then this field
'' contains the file name in the "8.3" form with a dot and without any extending blanks,
'' else the part of the long file name is saved in this field):
'' FFD_STRUCT_SHORT_SIZE + FFD_NAME_SIZE + FFD_EXT_SIZE + 1[for "dot"]
#define	FFD_STRUCT_ABRIDGED_SIZE   44

'' Full size of a file find data structure (saved normally in a string):
'' FFD_STRUCT_SHORT_SIZE + MAX_LONGNAME_SIZE
#define	FFD_STRUCT_FULL_SIZE	288

#define	FFD_STRUCT_DEFAULT_SIZE	FFD_STRUCT_ABRIDGED_SIZE


'' Size of the file index table (TOTAL_FILE_NUMBER defined in fs_conf.inc)
#define FILE_INDEX_TABLE_SIZE	TOTAL_FILE_NUMBER
'' Size of one file index
#define FILE_INDEX_SIZE		32
'' Size of the buffer for temporary file data
#define FILE_DATA_BUFFER_SIZE	512

#define ROOT_DIR		"\"	'' ".\"

#define	DEF_CURRENTDIR		ROOT_DIR

#endif

