''**************************************************************************
'' Implementation of subroutines working with SmartMedia and conforming to
'' the SmartMedia specifications and to the special features
'' of the SmartMedia device driver
''**************************************************************************

#ifndef FS_SMC_I_INC
#define FS_SMC_I_INC

'' protection against eraising for the CIS block
byte bIsCisProtected

long  lSmMakerCode		' <1> Maker-Code = Hersteller
long  lSmIdCode			' <1> ID-Code = Karten-Id (Speicher-Groesse)
long  lSmAdrBytesRdWr		' <1> ADR-Bytes in READ / WRITE:  3 oder 4
long  lSmAdrBytesBlkErase	' <1> ADR-Bytes in Block-ERASE: 2 oder 3
long  lSmAdrOffsBlkErase	' <1> Byte-Offset of ADR in Block-ERASE: typical = 1
long  lSmBytesInSpare		' <2> Nr. of Bytes in Spare-Field (normal: 16 Bytes)
long  lSmSpareAdrMask		' <2> SPARE-ADR Mask (normal: 0FH)
long  lSmBytesInPage		' <4> Nr. of DATA Bytes in a Page (without spare-field)
long  lSmPageMask		' <2> ADR-Mask for this Page  (==>  RAM1_BYTES_IN_PAGE - 1 )
long  lSmPageIMask		' <2> inverse ADR-Mask for this Page, zB:  1111 1110 0000 0000
long  lSmPagesInBlock		' <2> Nr. of Pages in a Block
long  lSmBytesInBlock		' <4> Nr. of DATA-Bytes in a Block
long  lSmBlockAdrIncs		' <4> Bloc-ADR Increment (i.d.R. identisch mit "RAM1_BYTES_IN_BLOCK")
long  lSmBlockMask		' <2> Block-ADR Maske
long  lSmBlockIMask		' <2> Block-ADR inverse Maske
long  lSmBlockAdrSrl		' <1> Block-ADR Shifts-Right:  ADR --> to Block-Number (2**14 = 16384 = 16KByte)
long  lSmNoOfBlocks		' <2> Total Number of "Blocks" in this SmartMedia Card
long  lSmAdrHighBlock		' <4> Base ADR of highest Block
long  lSmAdrEnd			' <4> End-ADR = 1.ADR hinter dem letzen Byte !!

long  lSmNoOfInvBlocks		' No-of-Invalid Blocks (total)

string sSmInvBlks$(SM_INV_BLOCKS_TAB_SIZE)	

string sSmPageBuffer$(SM_MAX_PAGE_SIZE)
string sSmPgSpareBuf$(SM_MAX_PAGE_WITH_SPARE_SIZE)
string sSmBlockBuffer$(SM_MAX_BLOCK_SIZE)
long lSmPageSizeWithSpare

string sSmFindBuffer$(SM_FIND_BUFFER_SIZE)

long lSmWritePrevHandle

array laSmLogToPhysZonesSize(SM_MAX_ZONE_NUM) of long
#ifdef SM_USE_FAST_LOG2PHYS_FUNCS
string sSmLogToPhys$(SM_LOG_TO_PHYS_TABLE_MAX_SIZE)
string sSmPhysToLog$(SM_PHYS_TO_LOG_TABLE_MAX_SIZE)
#else
array laSmLogToPhys(SM_MAX_ZONE_NUM, SM_MAX_LOG_BLOCK_NUM_IN_ZONE) of long
'' string sSmLogToPhysCopy$(16K)
#endif
long lSmLogToPhysNumOfZones
long lSmMaxZoneSizePhys, lSmMaxZoneSizeLog

long lSmLastWriteError

long lSmCisBlockAddr
string sSmCisField$(SM_MAX_PAGE_WITH_SPARE_SIZE)

'' this variable is assigned only after reading of file system configuration
'' so it's a kind of extern variable for this project file
long lDataAreaStart

#ifdef SM_WRITE_ON_BLOCK_SWITCH
long lSmInMemBlockStart
long lSmInMemBlockSize

sub vSMInitInMemVars
  lSmInMemBlockStart = 0
  lSmInMemBlockSize = 0
end

sub bSMIsAddrInMemBlock( long lpAddrToCheck; var byte bpvInMem )
  if lSmInMemBlockSize > 0 and ( lSmInMemBlockStart <= lpAddrToCheck and &
    lSmInMemBlockStart + lSmInMemBlockSize > lpAddrToCheck ) then
      bpvInMem = TRUE
  else
    bpvInMem = FALSE
  endif
end

sub vInMemBlockFlush()
  long llIndx, llAddr
  string slBackDoor$(180)
  llAddr = lSmInMemBlockStart
  if lSmInMemBlockSize > 0 then
    call vSMWriteBlock( llAddr )
#cm
#ifdef FS_SECURE_INIT
    slBackDoor$ = FS_STR_VARS_INIT_VALUE	'$JK
#endif	'' FS_SECURE_INIT
    llIndx = 0
    '' check lSmInMemBlockSize without spare fields,
    '' but write the pages with spare field
    while llIndx < lSmPagesInBlock and llIndx * lSmBytesInPage < lSmInMemBlockSize
      sSmPgSpareBuf$ = mid$( sSmBlockBuffer$, llIndx * lSmPageSizeWithSpare, lSmPageSizeWithSpare )
      '' in order to be able to write into the spare field (also, over position of invalid block flag)
      get #SMCARD, #0, #SMCARD_FUNC_STATUS, 153, slBackDoor$
      put #SMCARD, #llAddr, #SMCARD_FUNC_WRITE_ABS, sSmPgSpareBuf$
      llAddr = llAddr + lSmBytesInPage
      llIndx = llIndx + 1
    endwhile
#endcm
    call vSMInitInMemVars()
    sSmBlockBuffer$ = ""
  endif
end

sub vInMemBlockRegistry( long lpBlockStart, lpBlockSize )
  lSmInMemBlockStart = lpBlockStart
  lSmInMemBlockSize = lpBlockSize
end

sub vInMemBlockGetPage( long lpPageAddr, lpSizeToRead )
  long llIndx
  llIndx = mod( lpPageAddr, lSmBytesInBlock ) / lSmBytesInPage
  sSmPageBuffer$ = mid$( sSmBlockBuffer$, llIndx * lSmPageSizeWithSpare, lpSizeToRead )
end

sub lInMemBlockGetNum( long lpNumAddr; byte bpSize; var long lpvNum )
  long llAddrInMemBlock, llAddrInNormBlock
  llAddrInNormBlock = mod( lpNumAddr, lSmBytesInBlock )
  llAddrInMemBlock = llAddrInNormBlock / lSmBytesInPage * lSmPageSizeWithSpare +&
    mod( llAddrInNormBlock, lSmBytesInPage )
  lpvNum = nfroms( sSmBlockBuffer$, llAddrInMemBlock, bpSize )
end
#endif

sub bSMMount( var byte bpvRet )	' check and mount SmartMedia Card
  long llSMMounted
  string slBuf$

#ifdef FS_SECURE_INIT
  slBuf$ = FS_STR_VARS_INIT_VALUE	'$JK
#endif	'' FS_SECURE_INIT

  bpvRet = FALSE
''  dir_port 8,    10000000B		' All P8 Pins = Out,  P87 = In (R/-B)
''  out      8,7FH,01111111B		' All pins = hi!
  ' P0-P4,P6 Pins = Out
  dir_pin  8, 0, 0		
  dir_pin  8, 1, 0		
  dir_pin  8, 2, 0		
  dir_pin  8, 3, 0		
#ifndef SM_SETUP_DRIVER_FOR_TEC
  dir_pin  8, 4, 0
#endif
  dir_pin  8, 6, 0		

#ifndef SM_SETUP_DRIVER_FOR_TEC
  out      8,5FH,01011111B		' P0-P4,P6 Pins = hi!
#else
  out      8,5FH,01001111B		' P0-P3,P6 Pins = hi!
#endif

''  dir_port 7,    11110000B		' Bit-0...3 = Output
''  out      7,    00001111B, 0FH		' A0: P70=0,   Shutdown: P71=1

					'	P3-3	ACLK	ePorts    LEDs
					'	P3-4	DCLK	ePorts    LEDs
'	         76543210		'	P3-5	INE	ePorts	  DIP-Switch
''  dir_port 3,    11000111B		' Bit-3,4,5 = Output   P3-6 = Trigger-Input
''  out      3,    00111000B, 255		' <-- all "high"

  get #SMCARD, #0, #SMCARD_FUNC_RESET, 0, slBuf$	' <-- Ufunc = 4 =  R e s e t   Device
  get #SMCARD, #0, #SMCARD_FUNC_MOUNT, 0, llSMMounted	' 00	' <-- Ufunc =  2 = Mount S-Media

  lSmWritePrevHandle = SM_WRITE_BAD_HANDLE

  if llSMMounted = 0 then
    bpvRet = TRUE
  endif
end


sub lSMBitCount( long lpData; var long lpvCount )
  long llData
  lpvCount = 0
  llData = lpData
  while llData > 0
    lpvCount = lpvCount + ( llData bitand 01h )
    llData = llData shr 1
  endwhile
end


sub lSMLoadLogBlockAddr(var long lpvLogAddr)
  long llLogBlockAddr1, llLogBlockAddr2, llBitCount, llAddr
  llLogBlockAddr1 = &
    ( nfroms( sSmPgSpareBuf$, lSmBytesInPage + SM_SPARE_ADDR1_OFFS, 1 ) shl 8 ) bitor &
    nfroms( sSmPgSpareBuf$, lSmBytesInPage + SM_SPARE_ADDR1_OFFS + 1, 1 )
  llLogBlockAddr2 = &
    ( nfroms( sSmPgSpareBuf$, lSmBytesInPage + SM_SPARE_ADDR2_OFFS, 1 ) shl 8 ) bitor &
    nfroms( sSmPgSpareBuf$, lSmBytesInPage + SM_SPARE_ADDR2_OFFS + 1, 1 )
  if llLogBlockAddr1 = llLogBlockAddr2 then
    if ( llLogBlockAddr1 bitand SM_VALID_ADDR_MASK ) = SM_VALID_ADDR_ID then
      lpvLogAddr = ( llLogBlockAddr1 bitand SM_ADDR_CONTENTS_MASK ) shr 1
      return
    endif
  endif
  llAddr = llLogBlockAddr1 bitxor llLogBlockAddr2
  call lSMBitCount( llAddr, llBitCount )
  if llBitCount <> 01h then
    lpvLogAddr = ERROR
    return
  endif
  if ( llLogBlockAddr1 bitand SM_VALID_ADDR_MASK ) = SM_VALID_ADDR_ID then
    call lSMBitCount( llLogBlockAddr1, llBitCount )
    if ( llBitCount bitand 01h ) = 0 then
      lpvLogAddr = ( llLogBlockAddr1 bitand SM_ADDR_CONTENTS_MASK ) shr 1
      return
    endif
  endif
  if ( llLogBlockAddr2 bitand SM_VALID_ADDR_MASK ) = SM_VALID_ADDR_ID then
    call lSMBitCount( llLogBlockAddr2, llBitCount )
    if ( llBitCount bitand 01h ) = 0 then
      lpvLogAddr = ( llLogBlockAddr2 bitand SM_ADDR_CONTENTS_MASK ) shr 1
      return
    endif
  endif
  lpvLogAddr = ERROR
  return
end


sub vSMClearSpare()
  string slSpareFill$(SM_SPARE_MAX_SIZE)
#ifdef FS_SECURE_INIT
  slSpareFill$ = FS_STR_VARS_INIT_VALUE	'$JK
#endif	'' FS_SECURE_INIT
  slSpareFill$ = fill$( "FF"%, lSmBytesInSpare )
  sSmPgSpareBuf$ = stos$( sSmPgSpareBuf$, lSmBytesInPage, slSpareFill$, lSmBytesInSpare )
end


'' set the address of the logical block in the spare field of the physical page
sub vSMSetLogBlockAddr( long lpLogBlockAddr )
  long llAddrToWrite, llBitCount
  sSmPgSpareBuf$ = ntos$( sSmPgSpareBuf$, lSmBytesInPage + SM_SPARE_DATA_OFFS, 1, 0ffh )
  sSmPgSpareBuf$ = ntos$( sSmPgSpareBuf$, lSmBytesInPage + SM_SPARE_STAT_OFFS, 1, 0ffh )
  llAddrToWrite = lpLogBlockAddr * 2 + 1000h
  call lSMBitCount( llAddrToWrite, llBitCount )
  if mod( llBitCount, 2 ) > 0 then
    llAddrToWrite = llAddrToWrite + 1
  endif
  sSmPgSpareBuf$ = ntos$( sSmPgSpareBuf$, lSmBytesInPage + SM_SPARE_ADDR1_OFFS, 1, llAddrToWrite / 100h )
  sSmPgSpareBuf$ = ntos$( sSmPgSpareBuf$, lSmBytesInPage + SM_SPARE_ADDR2_OFFS, 1, llAddrToWrite / 100h )
  sSmPgSpareBuf$ = ntos$( sSmPgSpareBuf$, lSmBytesInPage + SM_SPARE_ADDR1_OFFS + 1, 1, llAddrToWrite )
  sSmPgSpareBuf$ = ntos$( sSmPgSpareBuf$, lSmBytesInPage + SM_SPARE_ADDR2_OFFS + 1, 1, llAddrToWrite )
end


sub bSMLogBlockToZoneIndexPair( long lpLogAddr; var long lpvZone, lpvIndexInZone; &
  var byte bpvIsCorrect )
  long llLogBlockIndex
  lpvZone = lpLogAddr / lSmMaxZoneSizeLog
  lpvIndexInZone = mod( lpLogAddr, lSmMaxZoneSizeLog ) / lSmBytesInBlock
  llLogBlockIndex = lpvZone * SM_MAX_LOG_BLOCK_NUM_IN_ZONE + lpvIndexInZone
  bpvIsCorrect = FALSE
  if llLogBlockIndex < lSmLogToPhysNumOfZones * SM_MAX_LOG_BLOCK_NUM_IN_ZONE then
    if lpvZone < lSmLogToPhysNumOfZones then
''    if lpvIndexInZone < laSmLogToPhysZonesSize(lpvZone) then
      if lpvIndexInZone < SM_MAX_LOG_BLOCK_NUM_IN_ZONE then
        bpvIsCorrect = TRUE
      endif
    endif
  endif
end


sub bSMIsLogBlockAssigned( long lpLogBlock; var byte bpvIsAssigned )
  byte blIsCorrect
  long llZone, llIndexInZone
  bpvIsAssigned = FALSE
#ifdef SM_USE_FAST_LOG2PHYS_FUNCS
  if nfroms( sSmLogToPhys$, lpLogBlock*SM_CONV_TAB_ENTRY_SIZE, SM_CONV_TAB_ENTRY_SIZE ) <> 0 then
    bpvIsAssigned = TRUE
  endif
#else  '' SM_USE_FAST_LOG2PHYS_FUNCS
  call bSMLogBlockToZoneIndexPair( lpLogBlock, llZone, llIndexInZone, blIsCorrect )
  if blIsCorrect = TRUE then
    if llIndexInZone < laSmLogToPhysZonesSize(llZone) then
      if laSmLogToPhys(llZone, llIndexInZone) <> NO_ASSIGN then
        bpvIsAssigned = TRUE
      endif
    endif
  endif
#endif  '' SM_USE_FAST_LOG2PHYS_FUNCS
end


'' find next free physical block
sub lSMFindFreePhysBlock( long lpStartBlockAddr; var long lpvFreeBlockAddr )
  long llNextBlockAddr, llBlockIndex, llZoneIndex, llDataAreaStartBlockPhys
#ifdef SM_USE_FAST_LOG2PHYS_FUNCS
  long llCurBlkPtr, llCurBlkSize, llPhysIndex, llPhysToLogLen, llOffset
  byte blIsFound, blIsSecond
  blIsFound = FALSE
  blIsSecond = FALSE
  llPhysToLogLen = len( sSmPhysToLog$ )
  llCurBlkPtr = lpStartBlockAddr / lSmBytesInBlock * SM_CONV_TAB_ENTRY_SIZE
  while blIsFound = FALSE and llCurBlkPtr < llPhysToLogLen
    llOffset = 0
    if llCurBlkPtr + 254 > llPhysToLogLen then
      llCurBlkSize = llPhysToLogLen - llCurBlkPtr
    else
      llCurBlkSize = 254
    endif
    sSmFindBuffer$ = mid$( sSmPhysToLog$, llCurBlkPtr, llCurBlkSize )
''    llLogIndex = chk_fnam( sSmFindBuffer$, 0, SM_CONV_TAB_ENTRY_SIZE, SM_UNASSIGNED_BLOCK_MASK_STR, 0 )
    while blIsFound = FALSE and llOffset < llCurBlkSize
      llPhysIndex = instr( SM_UNASSIGNED_BLOCK_MASK_STR, sSmFindBuffer$, llOffset, 4 )
      if llPhysIndex >= 0 then
        if mod( llPhysIndex, SM_CONV_TAB_ENTRY_SIZE ) <> 0 then
          llOffset = llPhysIndex + 1
        else
          if blIsSecond = FALSE then
            llOffset = llPhysIndex + SM_CONV_TAB_ENTRY_SIZE
            blIsSecond = TRUE
          else
            lpvFreeBlockAddr = ( ( llPhysIndex + llCurBlkPtr ) / SM_CONV_TAB_ENTRY_SIZE ) * lSmBytesInBlock
            blIsFound = TRUE
          endif
        endif
      else
        llOffset = llCurBlkSize
      endif
    endwhile
    llCurBlkPtr = llCurBlkPtr + llCurBlkSize
  endwhile
#else  '' SM_USE_FAST_LOG2PHYS_FUNCS
  llNextBlockAddr = lpStartBlockAddr
  call lSMConvLogToPhysAddr( lDataAreaStart, llDataAreaStartBlockPhys, SM_L2P_READ_ONLY )
  llDataAreaStartBlockPhys = llDataAreaStartBlockPhys / lSmBytesInBlock * lSmBytesInBlock
  if llNextBlockAddr < llDataAreaStartBlockPhys then
    llNextBlockAddr = llDataAreaStartBlockPhys
  endif
  lpvFreeBlockAddr = ERROR
  while lpvFreeBlockAddr = ERROR
    get #SMCARD, #llNextBlockAddr, #SMCARD_FUNC_READ_ABS, &
      lSmPageSizeWithSpare, sSmPgSpareBuf$
    '' is the current block valid
    if nfroms( sSmPgSpareBuf$, lSmBytesInPage + SM_SPARE_STAT_OFFS,&
      1 ) = SM_VALID_BLOCK_MARK then
        lpvFreeBlockAddr = llNextBlockAddr
        llZoneIndex = 0
        while llZoneIndex < lSmLogToPhysNumOfZones and lpvFreeBlockAddr = llNextBlockAddr
          llBlockIndex = 0
          while llBlockIndex < laSmLogToPhysZonesSize(llZoneIndex) and lpvFreeBlockAddr = llNextBlockAddr
            '' is the current block already used
            if lpvFreeBlockAddr = laSmLogToPhys(llZoneIndex, llBlockIndex) then
              lpvFreeBlockAddr = ERROR
            endif
            llBlockIndex = llBlockIndex + 1
          endwhile
          llZoneIndex = llZoneIndex + 1
        endwhile
    endif
    llNextBlockAddr = llNextBlockAddr + lSmBytesInBlock
  endwhile
#endif  '' SM_USE_FAST_LOG2PHYS_FUNCS
end


'' make new logical block
sub bSMMakeNewLogBlock( long lpZone, lpPosInZone; var byte bpvIsBlockMade )
  byte blIsBlockErased
  long llFreePhysBlockAddr, llNextPageInBlockAddr, llPageInBlockAddr, &
    llLogIndex, llPhysBlockIndex, llPhysPosInZone, llStartPhysBlockAddr
  bpvIsBlockMade = FALSE

  '' find next free physical block
  llStartPhysBlockAddr = lpZone * SM_MAX_PHYS_BLOCK_NUM_IN_ZONE * lSmBytesInBlock
  call lSMFindFreePhysBlock( llStartPhysBlockAddr, llFreePhysBlockAddr )

  if llFreePhysBlockAddr <> ERROR then

#ifdef SM_WRITE_ON_BLOCK_SWITCH
    call vInMemBlockFlush()
    call vInMemBlockRegistry( llFreePhysBlockAddr, lSmBytesInBlock )
#endif
    '' assign the address of the logical block to all pages of the physical block
    llPageInBlockAddr = 0
    for llNextPageInBlockAddr = llFreePhysBlockAddr to &
      llFreePhysBlockAddr + lSmBytesInBlock - lSmBytesInPage step lSmBytesInPage
        get #SMCARD, #llNextPageInBlockAddr, #SMCARD_FUNC_READ_ABS, &
      	  lSmPageSizeWithSpare, sSmPgSpareBuf$
    	call vSMSetLogBlockAddr( lpPosInZone )
  	'' Write the page to the block buffer
  	sSmBlockBuffer$ = stos$( sSmBlockBuffer$, llPageInBlockAddr, &
          sSmPgSpareBuf$, 0, lSmPageSizeWithSpare )
        llPageInBlockAddr = llPageInBlockAddr + lSmPageSizeWithSpare
    next

    call vSMWriteBlock( llFreePhysBlockAddr )

#ifdef SM_USE_FAST_LOG2PHYS_FUNCS
    llLogIndex = lpZone * SM_MAX_LOG_BLOCK_NUM_IN_ZONE + lpPosInZone
    llPhysBlockIndex = llFreePhysBlockAddr / lSmBytesInBlock
    llPhysPosInZone = mod( llFreePhysBlockAddr, lSmMaxZoneSizePhys ) / lSmBytesInBlock
    sSmLogToPhys$ = ntos$( sSmLogToPhys$, llLogIndex*SM_CONV_TAB_ENTRY_SIZE, SM_CONV_TAB_ENTRY_SIZE, llPhysPosInZone )
    sSmPhysToLog$ = ntos$( sSmPhysToLog$, llPhysBlockIndex*SM_CONV_TAB_ENTRY_SIZE, SM_CONV_TAB_ENTRY_SIZE, lpPosInZone )
#else  '' SM_USE_FAST_LOG2PHYS_FUNCS
    '' update the "LogToPhys" table
    laSmLogToPhys(lpZone, lpPosInZone) = llFreePhysBlockAddr
    if lpPosInZone >= laSmLogToPhysZonesSize(lpZone) then
      laSmLogToPhysZonesSize(lpZone) = lpPosInZone + 1
    endif
#endif  '' SM_USE_FAST_LOG2PHYS_FUNCS

    bpvIsBlockMade = TRUE
  endif
end


sub vSMSetNumOfZones()
  long llSmTotalBytes
  llSmTotalBytes = lSmBytesInBlock * lSmNoOfBlocks
  if llSmTotalBytes <= 16M then
    lSmLogToPhysNumOfZones = 1
  else
    if llSmTotalBytes <= 32M then
      lSmLogToPhysNumOfZones = 2
    else
      if llSmTotalBytes <= 64M then
        lSmLogToPhysNumOfZones = 4
      else
        if llSmTotalBytes <= 128M then
    	  lSmLogToPhysNumOfZones = 8
        endif
      endif
    endif
  endif
end

#cm
STRING BLK$ (16384), IBLK$ (16384)	'
  ADR = 0				'
  PUT #9, #ADR,  #22,  IBLK$		' Set inverse
  GET #9, #ADR,  #22,  0,BLK$		' read log-Block-ADRs
#endcm

sub vSMMakeLogToPhysTable()
#ifdef SM_USE_FAST_LOG2PHYS_FUNCS
  sSmLogToPhys$ = fill$( "the quick brown fox", 16k )
  sSmLogToPhys$ = "xx"
  sSmPhysToLog$ = ""
  put #SMCARD, #0,  #SMCARD_FUNC_READ_L2P_TABLE,  sSmLogToPhys$		' Set inverse
  ' take data-fields of 1000*SM_CONV_TAB_ENTRY_SIZE, and skip in-betweens of 24*SM_CONV_TAB_ENTRY_SIZE Byte
  get #SMCARD, #0,  #SMCARD_FUNC_READ_P2L_TABLE,  0, sSmPhysToLog$	' read log-Block-ADRs
''  concentrate$( sSmLogToPhys$, 0, SM_MAX_LOG_BLOCK_NUM_IN_ZONE * SM_CONV_TAB_ENTRY_SIZE, &
''    ( 1k - SM_MAX_LOG_BLOCK_NUM_IN_ZONE ) * SM_CONV_TAB_ENTRY_SIZE )
  call vSMSetNumOfZones()
  lSmMaxZoneSizePhys = SM_MAX_PHYS_BLOCK_NUM_IN_ZONE * lSmBytesInBlock
  lSmMaxZoneSizeLog = SM_MAX_LOG_BLOCK_NUM_IN_ZONE * lSmBytesInBlock
#else  '' SM_USE_FAST_LOG2PHYS_FUNCS
  byte blIsBlockErased
  long llNextBlockAddr, llLogBlockAddr, llBlockIndex, llAdjustedAddr, &
    llZoneIndex, llFinalPageAddr, llFinalPageLogBlockAddr'''', llTicks

''''  llTicks = ticks()

  llNextBlockAddr = 0

  call vSMSetNumOfZones()
  lSmMaxZoneSizePhys = SM_MAX_PHYS_BLOCK_NUM_IN_ZONE * lSmBytesInBlock
  lSmMaxZoneSizeLog = SM_MAX_LOG_BLOCK_NUM_IN_ZONE * lSmBytesInBlock

  for llZoneIndex = 0 to lSmLogToPhysNumOfZones - 1
    for llBlockIndex = 0 to SM_MAX_LOG_BLOCK_NUM_IN_ZONE - 1
      laSmLogToPhys(llZoneIndex, llBlockIndex) = NO_ASSIGN
    next
    laSmLogToPhysZonesSize(llZoneIndex) = 0
  next

  llZoneIndex = 0
  for llBlockIndex = 0 to lSmNoOfBlocks
    call lSMAdjustAddr( llNextBlockAddr, llAdjustedAddr )
    llNextBlockAddr = llAdjustedAddr
    get #SMCARD, #llNextBlockAddr, #SMCARD_FUNC_READ_ABS, &
      lSmPageSizeWithSpare, sSmPgSpareBuf$
    if nfroms( sSmPgSpareBuf$, lSmBytesInPage + SM_SPARE_STAT_OFFS, 1 ) = SM_VALID_BLOCK_MARK and &
      nfroms( sSmPgSpareBuf$, lSmBytesInPage + SM_SPARE_DATA_OFFS, 1 ) = SM_VALID_DATA_MARK then
        call lSMLoadLogBlockAddr( llLogBlockAddr )
        if llLogBlockAddr <> ERROR and llLogBlockAddr < SM_MAX_LOG_BLOCK_NUM_IN_ZONE then
          if laSmLogToPhys(llZoneIndex, llLogBlockAddr) = NO_ASSIGN then
            '' entry the address of the current physical block into the lookup table
            laSmLogToPhys(llZoneIndex, llLogBlockAddr) = llNextBlockAddr
            if llLogBlockAddr >= laSmLogToPhysZonesSize(llZoneIndex) then
              laSmLogToPhysZonesSize(llZoneIndex) = llLogBlockAddr + 1
            endif
''            sSmLogToPhysCopy$ = ntos$( sSmLogToPhysCopy$, llZoneIndex * 1024 + llLogBlockAddr*2, 2,&
''              llNextBlockAddr / lSmBytesInBlock )
          else
            '' read the final page of the current physical block
            llFinalPageAddr = llNextBlockAddr + lSmBytesInBlock - lSmBytesInPage
            get #SMCARD, #llFinalPageAddr, #SMCARD_FUNC_READ_ABS, &
              lSmPageSizeWithSpare, sSmPgSpareBuf$
            if nfroms( sSmPgSpareBuf$, lSmBytesInPage + SM_SPARE_STAT_OFFS,&
              1 ) = SM_VALID_BLOCK_MARK then
              call lSMLoadLogBlockAddr( llFinalPageLogBlockAddr )
              '' is the log addr in the final page the same as in the first page
              if llFinalPageLogBlockAddr = llLogBlockAddr then
            	'' read the final page of the physical block already entried
                '' into the lookup table
            	llFinalPageAddr = laSmLogToPhys(llZoneIndex, llLogBlockAddr) + &
                  lSmBytesInBlock - lSmBytesInPage
            	get #SMCARD, #llFinalPageAddr, #SMCARD_FUNC_READ_ABS, &
              	  lSmPageSizeWithSpare, sSmPgSpareBuf$
            	if nfroms( sSmPgSpareBuf$, lSmBytesInPage + SM_SPARE_STAT_OFFS,&
              	  1 ) = SM_VALID_BLOCK_MARK then
              	  call lSMLoadLogBlockAddr( llFinalPageLogBlockAddr )
                  '' is the log addr in the final page the same as in the first page
                  if llFinalPageLogBlockAddr = llLogBlockAddr then
                    '' erase the current physical block
                    call bSMEraseBlock( llNextBlockAddr, blIsBlockErased )
                  else
                    '' erase the entried physical block
                    call bSMEraseBlock( laSmLogToPhys(llZoneIndex, llLogBlockAddr), blIsBlockErased )
                    '' entry the address of the current physical block into the lookup table
                    laSmLogToPhys(llZoneIndex, llLogBlockAddr) = llNextBlockAddr
                    if llLogBlockAddr >= laSmLogToPhysZonesSize(llZoneIndex) then
                      laSmLogToPhysZonesSize(llZoneIndex) = llLogBlockAddr + 1
                    endif
                  endif
                endif
              else
                '' erase the current physical block
                call bSMEraseBlock( llNextBlockAddr, blIsBlockErased )
              endif
            endif
          endif
        endif
    endif
    llNextBlockAddr = llNextBlockAddr + lSmBytesInBlock
    if mod( llNextBlockAddr, lSmMaxZoneSizePhys ) = 0 then
      llZoneIndex = llZoneIndex + 1
    endif
  next

''''  llTicks = diff_ticks(llTicks)
''''  call vDebugPrintStr( "SMMakeLogToPhysTable: ", FALSE )
''''  call vDebugPrintNum( llTicks, TRUE )
#endif  '' SM_USE_FAST_LOG2PHYS_FUNCS
end


sub lSMConvLogToPhysAddr( long lpLogAddr; var long lpvPhysAddr; byte bpType )
  byte blIsBlockAvailable, blIsCorrect
  long llZone, llIndexInZone, llLogBlockIndex
#ifdef SM_USE_FAST_LOG2PHYS_FUNCS
  blIsBlockAvailable = TRUE
  call bSMLogBlockToZoneIndexPair( lpLogAddr, llZone, llIndexInZone, blIsCorrect )
  if blIsCorrect = TRUE then
    llLogBlockIndex = lpLogAddr / lSmBytesInBlock
    lpvPhysAddr = nfroms( sSmLogToPhys$, llLogBlockIndex*SM_CONV_TAB_ENTRY_SIZE, SM_CONV_TAB_ENTRY_SIZE )
    if lpvPhysAddr = SM_UNASSIGNED_BLOCK_MASK then
      if bpType = SM_L2P_CREATE then
        call bSMMakeNewLogBlock( llZone, llIndexInZone, blIsBlockAvailable )
      else
        blIsBlockAvailable = FALSE
      endif
    endif
    if blIsBlockAvailable = TRUE then
      lpvPhysAddr = nfroms( sSmLogToPhys$, llLogBlockIndex*SM_CONV_TAB_ENTRY_SIZE, SM_CONV_TAB_ENTRY_SIZE )
      lpvPhysAddr = llZone * lSmMaxZoneSizePhys + lpvPhysAddr * lSmBytesInBlock
      lpvPhysAddr = lpvPhysAddr + mod( lpLogAddr, lSmBytesInBlock )
    endif
  endif
#else  '' SM_USE_FAST_LOG2PHYS_FUNCS
  lpvPhysAddr = NO_ASSIGN
  call bSMLogBlockToZoneIndexPair( lpLogAddr, llZone, llIndexInZone, blIsCorrect )
  if blIsCorrect = TRUE then
    blIsBlockAvailable = TRUE
    if laSmLogToPhys(llZone, llIndexInZone) = NO_ASSIGN then
      if bpType = SM_L2P_CREATE then
        call bSMMakeNewLogBlock( llZone, llIndexInZone, blIsBlockAvailable )
      else
        blIsBlockAvailable = FALSE
      endif
    endif
    if blIsBlockAvailable = TRUE then
      lpvPhysAddr = laSmLogToPhys(llZone, llIndexInZone) + mod( lpLogAddr, lSmBytesInBlock )
    endif
  endif
#endif  '' SM_USE_FAST_LOG2PHYS_FUNCS
end


sub bSMLoadCisField( var byte bpvIsCisOk )
  long llNextBlockAddr, llBlockIndex
  bpvIsCisOk = FALSE
  lSmCisBlockAddr = -1
  llNextBlockAddr = 0
  for llBlockIndex = 0 to lSmNoOfBlocks
    get #SMCARD, #llNextBlockAddr, #SMCARD_FUNC_READ_ABS, &
      lSmPageSizeWithSpare, sSmPgSpareBuf$
    if nfroms( sSmPgSpareBuf$, lSmBytesInPage + SM_SPARE_STAT_OFFS,&
      1 ) = SM_VALID_BLOCK_MARK then
        if mid$( sSmPgSpareBuf$, SM_CIS_ID1_OFFS, SM_CIS_ID_LEN ) =&
          mid$( sSmPgSpareBuf$, SM_CIS_ID2_OFFS, SM_CIS_ID_LEN ) then
            if mid$( sSmPgSpareBuf$, SM_CIS_ID1_OFFS, SM_CIS_ID_LEN ) = SM_CIS_ID_PATTERN then
              lSmCisBlockAddr = llNextBlockAddr
	      bpvIsCisOk = TRUE
              return
            endif
        endif
    endif
    llNextBlockAddr = llNextBlockAddr + lSmBytesInBlock
  next
end


sub bSMInitCisField( var byte bpvCisFieldOk )
  bpvCisFieldOk = FALSE
  if lSmBytesInPage = 512 then
    sSmCisField$ = &
    	"01 03 d9 01 ff 18 02 df 01 20 04 00 00 00 00 21 &
         02 04 01 22 02 01 01 22 03 02 04 07 1a 05 01 03 &
         00 02 0f 1b 08 c0 c0 a1 01 55 08 00 20 1b 0a c1 &
         41 99 01 55 64 f0 ff ff 20 1b 0c 82 41 18 ea 61 &
         f0 01 07 f6 03 01 ee 1b 0c 83 41 18 ea 61 70 01 &
         07 76 03 01 ee 15 14 05 00 20 20 20 20 20 20 20 &
         00 20 20 20 20 00 30 2e 30 00 ff 14 00 ff 00 00 &
         00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 &
         00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 &
         00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 &
         00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 &
         00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 &
         00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 &
         00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 &
         00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 &
         00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 &
    	 01 03 d9 01 ff 18 02 df 01 20 04 00 00 00 00 21 &
         02 04 01 22 02 01 01 22 03 02 04 07 1a 05 01 03 &
         00 02 0f 1b 08 c0 c0 a1 01 55 08 00 20 1b 0a c1 &
         41 99 01 55 64 f0 ff ff 20 1b 0c 82 41 18 ea 61 &
         f0 01 07 f6 03 01 ee 1b 0c 83 41 18 ea 61 70 01 &
         07 76 03 01 ee 15 14 05 00 20 20 20 20 20 20 20 &
         00 20 20 20 20 00 30 2e 30 00 ff 14 00 ff 00 00 &
         00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 &
         00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 &
         00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 &
         00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 &
         00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 &
         00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 &
         00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 &
         00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 &
         00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 &
         ff ff ff ff ff ff 00 00 0c cc c3 00 00 0c cc c3"%

    bpvCisFieldOk = TRUE
  endif
end


sub vSMSetCisField()
  byte blBlockIndexReached, blCisFieldInitOk, blIsBlockErased, blCurByte
  long llBlockAddr, llBlockIndex, llIndx, llBitIndx

  call bSMInitCisField( blCisFieldInitOk )
  if blCisFieldInitOk = FALSE then
    return
  endif
  blBlockIndexReached = FALSE
  llBlockIndex = 0
  llIndx = 0
  while blBlockIndexReached = FALSE and llIndx < lSmNoOfInvBlocks / 8
    blCurByte = nfroms( sSmInvBlks$, llIndx, 1 )
    llBitIndx = 0
    while blBlockIndexReached = FALSE and llBitIndx < 8
      if ( blCurByte shr llBitIndx ) bitand 1 = 1 then
        llBlockIndex = llBlockIndex + 1
      else
        blBlockIndexReached = TRUE
      endif
      llBitIndx = llBitIndx + 1
    endwhile
    llIndx = llIndx + 1
  endwhile
  llBlockAddr = llBlockIndex * lSmBytesInBlock
  sSmBlockBuffer$ = fill$( "<00>", SM_MAX_BLOCK_SIZE )
  sSmBlockBuffer$ = stos$( sSmBlockBuffer$, 0, sSmCisField$, 0, len(sSmCisField$) )
  call vSMWriteBlock( llBlockAddr )
end


sub bSMEraseBlock( long lpAddr; var byte bpvIsOk )
  string slBlockState$(1)
  byte blSmStat
  bpvIsOk = FALSE
  blSmStat = 0

#ifdef FS_SECURE_INIT
  slBlockState$ = FS_STR_VARS_INIT_VALUE	'$JK
#endif	'' FS_SECURE_INIT

  if bIsCisProtected = TRUE then
    if lpAddr = lSmCisBlockAddr then
      return
    endif
  endif
  get #SMCARD, #lpAddr, #SMCARD_FUNC_ERASE, 0, slBlockState$	' <-- Ufunc =  17 = Erase Block
  while (blSmStat bitand SMCARD_ERASE_BUSY_FLAG) <> SMCARD_ERASE_BUSY_FLAG
    get #SMCARD, #0, #SMCARD_FUNC_STATUS, 1, blSmStat	' C0	' <-- Ufunc =  5 = Read-Flash-Status
  endwhile
  if (blSmStat bitand SMCARD_ERASE_BUSY_FLAG) = SMCARD_ERASE_BUSY_FLAG then
    bpvIsOk = TRUE
  endif
end


sub bSMGetInfoSet( var string spvInfoSet$; var byte bpvIsRead )
  bpvIsRead = TRUE
  if max_len(spvInfoSet$) < SMCARD_INFO_BLOCK_SIZE then
    bpvIsRead = FALSE
  else
    get #SMCARD, #0, #SMCARD_FUNC_INFO, 0, spvInfoSet$	' Get ID + Infoset
  endif
end


sub vSMReadConfiguration()			'
  long llSmResult				' Result-Flag of Smart-Media Operation
  string slSmId$( SMCARD_INFO_BLOCK_SIZE )	' ID of the Smart-Media:
  long llL

#ifdef FS_SECURE_INIT
  slSmId$ = FS_STR_VARS_INIT_VALUE	'$JK
#endif	'' FS_SECURE_INIT

  get #SMCARD, #0,    #SMCARD_FUNC_STATUS,    0,llSmResult' C0	' Read-Flash-Status
  get #SMCARD, #0,    #SMCARD_FUNC_INFO,    0,slSmId$		' Get ID + Infoset
  get #SMCARD, #0,    #SMCARD_FUNC_INV_BLOCK_TAB, 0,sSmInvBlks$	' Get invalid Block-Tabl

  lSmMakerCode = 	nfroms( slSmId$, SM_MAKER_CODE_POS, SM_MAKER_CODE_SIZE ) 	' <1> Maker-Code = Manufacturer
  lSmIdCode = 		nfroms( slSmId$, SM_ID_CODE_POS, SM_ID_CODE_SIZE ) 		' <1> ID-Code = Card-Id (Memory Size)
  lSmAdrBytesRdWr = 	nfroms( slSmId$, SM_ADR_BYTES_RD_WR_POS, SM_ADR_BYTES_RD_WR_SIZE ) 	' <1> ADR-Bytes in READ / WRITE:  3 or 4
  lSmAdrBytesBlkErase = nfroms( slSmId$, SM_ADR_BYTES_BLKERA_POS, SM_ADR_BYTES_BLKERA_SIZE ) 	' <1> ADR-Bytes in Block-ERASE: 2 or 3
  lSmAdrOffsBlkErase = 	nfroms( slSmId$, SM_ADR_OFFS_BLKERA_POS, SM_ADR_OFFS_BLKERA_SIZE ) 	' <1> Byte-Offset of ADR in Block-ERASE: typical = 1
  lSmBytesInSpare = 	nfroms( slSmId$, SM_BYTES_IN_SPARE_POS, SM_BYTES_IN_SPARE_SIZE ) 	' <2> Nr. of Bytes in Spare-Field (normal: 16 Bytes)
  lSmSpareAdrMask = 	nfroms( slSmId$, SM_SPARE_ADR_MASK_POS,  SM_SPARE_ADR_MASK_SIZE ) 	' <2> SPARE-ADR Mask (normal: 0FH)
  lSmBytesInPage = 	nfroms( slSmId$, SM_BYTES_IN_PAGE_POS, SM_BYTES_IN_PAGE_SIZE ) 	' <4> Nr. of DATA Bytes in a Page (without spare-field)
  lSmPageMask = 	nfroms( slSmId$, SM_PAGE_MASK_POS, SM_PAGE_MASK_SIZE ) 		' <2> ADR-Mask for this Page  (==>  RAM1_BYTES_IN_PAGE - 1 )
  lSmPageIMask = 	nfroms( slSmId$, SM_PAGE_IMASK_POS, SM_PAGE_IMASK_SIZE ) 	' <2> inverse ADR-Mask for this Page, f.e.:  1111 1110 0000 0000
  lSmPagesInBlock = 	nfroms( slSmId$, SM_PAGES_IN_BLOCK_POS, SM_PAGES_IN_BLOCK_SIZE )' <2> Nr. of Pages in a Block
  lSmBytesInBlock = 	nfroms( slSmId$, SM_BYTES_IN_BLOCK_POS, SM_BYTES_IN_BLOCK_SIZE )' <4> Nr. of DATA-Bytes in a Block
  lSmBlockAdrIncs = 	nfroms( slSmId$, SM_BLOCK_ADR_INCS_POS, SM_BLOCK_ADR_INCS_SIZE )' <4> Bloc-ADR Increment (as a rule identical with "RAM1_BYTES_IN_BLOCK")
  lSmBlockMask = 	nfroms( slSmId$, SM_BLOCK_MASK_POS, SM_BLOCK_MASK_SIZE ) 	' <2> Block-ADR Mask
  lSmBlockIMask = 	nfroms( slSmId$, SM_BLOCK_IMASK_POS, SM_BLOCK_IMASK_SIZE ) 	' <2> Block-ADR invers Mask
  lSmBlockAdrSrl = 	nfroms( slSmId$, SM_BLOCK_ADR_SRL_POS, SM_BLOCK_ADR_SRL_SIZE ) 	' <1> Block-ADR Shifts-Right:  ADR --> to Block-Number (2**14 = 16384 = 16KByte)
  lSmNoOfBlocks = 	nfroms( slSmId$, SM_NO_OF_BLOCKS_POS, SM_NO_OF_BLOCKS_SIZE ) 	' <2> Total Number of "Blocks" in this SmartMedia Card
  lSmAdrHighBlock = 	nfroms( slSmId$, SM_ADR_HIGH_BLOCK_POS, SM_ADR_HIGH_BLOCK_SIZE )' <4> Base ADR of highest Block
  lSmAdrEnd = 		nfroms( slSmId$, SM_ADR_END_POS, SM_ADR_END_SIZE ) 		' <4> End-ADR = 1.ADR after the last Byte !!

  llL = len( sSmInvBlks$ ) - 2
  lSmNoOfInvBlocks = 	nfroms( cut_and_paste$( sSmInvBlks$, llL, 2 ), 0, 2 )' Total Number of "Invalid-Blocks"
  lSmPageSizeWithSpare = lSmBytesInPage + lSmBytesInSpare
end


sub lSMGetMediaSize( var long lpvMediaSize )
  lpvMediaSize = lSmNoOfBlocks * lSmBytesInBlock
end


sub lSMAdjustAddr( long lpCurAddr; var long lpvNewAddr )
  lpvNewAddr = bit_map_adjust( sSmInvBlks$, lSmBytesInBlock, lpCurAddr, 1 )
''  if lpvNewAddr <> lpCurAddr then
''    lpvNewAddr = lpvNewAddr + mod( lpCurAddr, lSmBytesInBlock )
''  endif
end


sub lSMReadNum( long lpAddr; byte bpSize; var long lpvNum; var byte bpvIsOk )
  long llPhysAddr
  bpvIsOk = TRUE
  call lSMConvLogToPhysAddr( lpAddr, llPhysAddr, SM_L2P_READ_ONLY )
  if llPhysAddr = NO_ASSIGN then
    bpvIsOk = FALSE
    return
  endif
  call lSMReadNumPhys( llPhysAddr, bpSize, lpvNum )
end


sub lSMReadNumPhys( long lpPhysAddr; byte bpSize; var long lpvNum )
  byte blIsIn
  long llAdjustedAddr, llPageAddr, llPhysAddr
  string slBuffer$(SM_MAX_PAGE_SIZE)

#ifdef FS_SECURE_INIT
  slBuffer$ = FS_STR_VARS_INIT_VALUE	'$JK
#endif	'' FS_SECURE_INIT

  call lSMAdjustAddr( lpPhysAddr, llAdjustedAddr )
#ifdef SM_WRITE_ON_BLOCK_SWITCH
  call bSMIsAddrInMemBlock( llAdjustedAddr, blIsIn )
  if blIsIn = TRUE then
    call lInMemBlockGetNum( llAdjustedAddr, bpSize, lpvNum )
  else
#endif
    llPageAddr = llAdjustedAddr - mod( llAdjustedAddr, lSmBytesInPage )
    get #SMCARD, #llPageAddr, #SMCARD_FUNC_READ_ABS, SM_MAX_PAGE_SIZE, slBuffer$
    lpvNum = nfroms( slBuffer$, llAdjustedAddr - llPageAddr, bpSize )
#ifdef SM_WRITE_ON_BLOCK_SWITCH
  endif
#endif
end


sub sSMReadString( long lpAddr; long lpSize; var string spvBuffer$; var byte bpvIsOk )
  long llPhysAddr
  bpvIsOk = TRUE
  call lSMConvLogToPhysAddr( lpAddr, llPhysAddr, SM_L2P_READ_ONLY )
  if llPhysAddr = NO_ASSIGN then
    bpvIsOk = FALSE
    return
  endif
  call sSMReadStringPhys( llPhysAddr, lpAddr, lpSize, spvBuffer$ )
end


sub sSMReadStringPhys( long lpPhysAddr, lpLogAddr; long lpSize; var string spvBuffer$ )
  long llReadCount, llNextAddr, llActSize, llNextAdjustedAddr
  long llNextPageAddr, llPageReadCount, llOffsetInPage, llNextLogAddr
  byte blIsIn

  llActSize = lpSize
  llNextLogAddr = lpLogAddr
  llNextAddr = lpPhysAddr
  if mod( llNextAddr, lSmBytesInPage ) <> 0 then
    llReadCount = lSmBytesInPage - mod( llNextAddr, lSmBytesInPage )
    if llActSize < llReadCount then
      llReadCount = llActSize
    endif
    call lSMAdjustAddr( llNextAddr, llNextAdjustedAddr )
    '' now we can only read by pages
    llNextPageAddr = llNextAdjustedAddr - mod( llNextAdjustedAddr, lSmBytesInPage )
    llOffsetInPage = llNextAdjustedAddr - llNextPageAddr
    llPageReadCount = llOffsetInPage + llReadCount
#ifdef SM_WRITE_ON_BLOCK_SWITCH
    call bSMIsAddrInMemBlock( llNextPageAddr, blIsIn )
    if blIsIn = TRUE then
      call vInMemBlockGetPage( llNextPageAddr, llPageReadCount )
    else
#endif
      get #SMCARD, #llNextPageAddr, #SMCARD_FUNC_READ_ABS, llPageReadCount, sSmPageBuffer$
#ifdef SM_WRITE_ON_BLOCK_SWITCH
    endif
#endif
    spvBuffer$ = stos$( spvBuffer$, llNextAddr - lpPhysAddr, sSmPageBuffer$, llOffsetInPage, llReadCount )
    llNextAddr = llNextAddr + llReadCount
    llActSize = llActSize - llReadCount
  endif
  while llActSize > 0
    if llActSize > lSmBytesInPage then
      llReadCount = lSmBytesInPage
    else
      llReadCount = llActSize
    endif
    call lSMAdjustAddr( llNextAddr, llNextAdjustedAddr )
    '' now we can only read by pages
    llNextPageAddr = llNextAdjustedAddr - mod( llNextAdjustedAddr, lSmBytesInPage )
    llOffsetInPage = llNextAdjustedAddr - llNextPageAddr
    llPageReadCount = llOffsetInPage + llReadCount
#ifdef SM_WRITE_ON_BLOCK_SWITCH
    call bSMIsAddrInMemBlock( llNextPageAddr, blIsIn )
    if blIsIn = TRUE then
      call vInMemBlockGetPage( llNextPageAddr, llPageReadCount )
    else
#endif
      get #SMCARD, #llNextPageAddr, #SMCARD_FUNC_READ_ABS, llPageReadCount, sSmPageBuffer$
#ifdef SM_WRITE_ON_BLOCK_SWITCH
    endif
#endif
    spvBuffer$ = stos$( spvBuffer$, llNextAddr - lpPhysAddr, sSmPageBuffer$, llOffsetInPage, llReadCount )
    llNextAddr = llNextAddr + llReadCount
    llActSize = llActSize - llReadCount
    '' is the physical/logical block boundary reached
    if llNextLogAddr / lSmBytesInBlock < (llNextLogAddr + llReadCount) / lSmBytesInBlock then
      llNextLogAddr = llNextLogAddr + llReadCount
      call lSMConvLogToPhysAddr( llNextLogAddr, llNextAddr, SM_L2P_READ_ONLY )
      if llNextAddr = NO_ASSIGN then
        lSmLastWriteError = SM_RD_ERR_ADDR_NOT_ASSIGNED
        return
      endif
    else
      llNextLogAddr = llNextLogAddr + llReadCount
    endif
  endwhile
end


sub vSMWriteStringToMemPage( long lpAddr; long lpPos, lpSize; var string spvBuffer$ )
  long llPageInBlockAddr, llInPageAddr, llECCDuration
  sSmPgSpareBuf$ = ""
  '' Calculate the relative address in the page
  llInPageAddr = mod( lpAddr, lSmBytesInPage )
  '' Calculate the address of the page in block
  llPageInBlockAddr = mod( lpAddr / lSmBytesInPage, lSmPagesInBlock )
  llPageInBlockAddr = llPageInBlockAddr * lSmPageSizeWithSpare
  '' Read the previous contains of the page from block
  sSmPgSpareBuf$ = mid$( sSmBlockBuffer$, llPageInBlockAddr, lSmPageSizeWithSpare )
  '' Write new data to the page
  sSmPgSpareBuf$ = stos$( sSmPgSpareBuf$, llInPageAddr, spvBuffer$, lpPos, lpSize )
  '' Calculate ECC for data in "sSmPgSpareBuf$"
  call vSMUpdateECC()
  '' Write the page to the block buffer
  sSmBlockBuffer$ = stos$( sSmBlockBuffer$, llPageInBlockAddr, &
    sSmPgSpareBuf$, 0, lSmPageSizeWithSpare )
end


sub vSMWriteBlock( long lpPhysBlockAddr )
  byte blIsErased
  long llAddr, llIndx
  string slBackDoor$(180)

#ifdef FS_SECURE_INIT
  slBackDoor$ = FS_STR_VARS_INIT_VALUE	'$JK
#endif	'' FS_SECURE_INIT

  '' Erase the block
  blIsErased = TRUE
  call bSMEraseBlock( lpPhysBlockAddr, blIsErased )
  if blIsErased = FALSE then
    lSmLastWriteError = SM_WR_ERR_BLOCK_NOT_ERASED
    return
  endif
  llAddr = lpPhysBlockAddr
  for llIndx = 0 to lSmPagesInBlock - 1
    sSmPgSpareBuf$ = mid$( sSmBlockBuffer$, llIndx * lSmPageSizeWithSpare, lSmPageSizeWithSpare )
    '' in order to be able to write into the spare field (also, over position of invalid block flag)
    get #SMCARD, #0, #SMCARD_FUNC_STATUS, 153, slBackDoor$
    put #SMCARD, #llAddr, #SMCARD_FUNC_WRITE_ABS, sSmPgSpareBuf$
    llAddr = llAddr + lSmBytesInPage
  next
end


sub vSMWriteString( long lpAddr, lpPos, lpSize; var string spvBuffer$ )
  long llWrittenCount, llNextAddr, llActSize, llBlockAddr, llNextAdjustedAddr, llPhysAddr
  long llPos, llIndx, llAddr, llNextLogAddr
  byte blIsErased, blIsIn

  call lSMConvLogToPhysAddr( lpAddr, llPhysAddr, SM_L2P_CREATE )
  if llPhysAddr = NO_ASSIGN then
    lSmLastWriteError = SM_WR_ERR_ADDR_NOT_ASSIGNED
    return
  endif
  llNextAddr = llPhysAddr
  llWrittenCount = 0
  llBlockAddr = -1
  call lSMAdjustAddr( llNextAddr, llNextAdjustedAddr )

#ifdef SM_WRITE_ON_BLOCK_SWITCH
  '' if the next address doesn't belong to the block previously saved in memory,
  '' write the saved block to the media physically
  if lpSize > 0 then
    call bSMIsAddrInMemBlock( llNextAdjustedAddr, blIsIn )
    if blIsIn = FALSE then
      call vInMemBlockFlush()
    endif
  endif
#else
  sSmBlockBuffer$ = ""
#endif

  while llWrittenCount < lpSize
    if llBlockAddr = -1 or llNextAdjustedAddr >= llBlockAddr + lSmBytesInBlock then
      if llBlockAddr <> -1 then
        '' Write previous block to disk
#ifdef SM_WRITE_ON_BLOCK_SWITCH
	call vInMemBlockFlush()
#else
	call vSMWriteBlock( llBlockAddr )
#endif
  	llNextLogAddr = lpAddr + llWrittenCount
        call lSMConvLogToPhysAddr( llNextLogAddr, llNextAddr, SM_L2P_CREATE )
  	if llNextAddr = NO_ASSIGN then
    	  lSmLastWriteError = SM_WR_ERR_ADDR_NOT_ASSIGNED
          return
  	endif
  	call lSMAdjustAddr( llNextAddr, llNextAdjustedAddr )
      endif
      llBlockAddr = llNextAdjustedAddr / lSmBytesInBlock * lSmBytesInBlock
#ifdef SM_WRITE_ON_BLOCK_SWITCH
      call bSMIsAddrInMemBlock( llNextAdjustedAddr, blIsIn )
      if blIsIn = FALSE then
#endif
      	'' Read the block to which "llNextAddr" belongs
      	llAddr = llBlockAddr
      	for llIndx = 0 to lSmPagesInBlock - 1
          get #SMCARD, #llAddr, #SMCARD_FUNC_READ_ABS, lSmPageSizeWithSpare, sSmPgSpareBuf$
          sSmBlockBuffer$ = stos$( sSmBlockBuffer$, llIndx * lSmPageSizeWithSpare, &
            sSmPgSpareBuf$, 0, lSmPageSizeWithSpare )
          llAddr = llAddr + lSmBytesInPage
      	next
#ifdef SM_WRITE_ON_BLOCK_SWITCH
        call vInMemBlockRegistry( llBlockAddr, lSmBytesInBlock )
#endif
''''        '' Erase the block
''''       	call bSMEraseBlock( llBlockAddr, blIsErased )
''''      	if blIsErased = FALSE then
''''          lSmLastWriteError = SM_WR_ERR_BLOCK_NOT_ERASED
''''          return
''''      	endif
#ifdef SM_WRITE_ON_BLOCK_SWITCH
      endif
#endif
    endif
    if lpSize - llWrittenCount < lSmBytesInPage then
      llActSize = lpSize - llWrittenCount
    else
      '' "llNextAddr" is in the middle of a page
      if mod( llNextAddr, lSmBytesInPage ) > 0 then
        llActSize = lSmBytesInPage - mod( llNextAddr, lSmBytesInPage )
      else
        llActSize = lSmBytesInPage
      endif
    endif

    llPos = lpPos + llWrittenCount
    call vSMWriteStringToMemPage( llNextAddr, llPos, llActSize, spvBuffer$ )

    llWrittenCount = llWrittenCount + llActSize
    llNextAddr = llNextAddr + llActSize
    call lSMAdjustAddr( llNextAddr, llNextAdjustedAddr )
  endwhile

  '' Write last block to disk
#ifndef SM_WRITE_ON_BLOCK_SWITCH
  if lpSize > 0 then
    call vSMWriteBlock( llBlockAddr )
  endif
#endif
end


sub bSMInit( var byte bpvRet )
  byte blSMMounted, blIsCisOk
  bpvRet = FALSE
#ifdef FS_SECURE_INIT
  call vSMInitGlobalVars()
#endif	'' FS_SECURE_INIT
  call bSMMount( blSMMounted )
  if blSMMounted = TRUE then
    call vSMReadConfiguration()
    call bSMLoadCisField( blIsCisOk )
    if blIsCisOk = TRUE then
      call vSMInitECCTable()
#ifdef SM_WRITE_ON_BLOCK_SWITCH
      call vSMInitInMemVars()
#endif
      call vSMMakeLogToPhysTable()
      bIsCisProtected = TRUE
      bpvRet = TRUE
    endif
  endif
end


#ifdef FS_SECURE_INIT
sub vSMInitGlobalVars()
  word wlIndx, wlSecIndx
  bIsCisProtected = TRUE	' protection against eraising for the CIS block

  lSmMakerCode = FS_NUM_VARS_INIT_VALUE		' <1> Maker-Code = Hersteller
  lSmIdCode = FS_NUM_VARS_INIT_VALUE	  	' <1> ID-Code = Karten-Id (Speicher-Groesse)
  lSmAdrBytesRdWr = FS_NUM_VARS_INIT_VALUE	' <1> ADR-Bytes in READ / WRITE:  3 oder 4
  lSmAdrBytesBlkErase = FS_NUM_VARS_INIT_VALUE	' <1> ADR-Bytes in Block-ERASE: 2 oder 3
  lSmAdrOffsBlkErase = FS_NUM_VARS_INIT_VALUE	' <1> Byte-Offset of ADR in Block-ERASE: typical = 1
  lSmBytesInSpare = FS_NUM_VARS_INIT_VALUE	' <2> Nr. of Bytes in Spare-Field (normal: 16 Bytes)
  lSmSpareAdrMask = FS_NUM_VARS_INIT_VALUE	' <2> SPARE-ADR Mask (normal: 0FH)
  lSmBytesInPage = FS_NUM_VARS_INIT_VALUE	' <4> Nr. of DATA Bytes in a Page (without spare-field)
  lSmPageMask = FS_NUM_VARS_INIT_VALUE		' <2> ADR-Mask for this Page  (==>  RAM1_BYTES_IN_PAGE - 1 )
  lSmPageIMask = FS_NUM_VARS_INIT_VALUE		' <2> inverse ADR-Mask for this Page, zB:  1111 1110 0000 0000
  lSmPagesInBlock = FS_NUM_VARS_INIT_VALUE	' <2> Nr. of Pages in a Block
  lSmBytesInBlock = FS_NUM_VARS_INIT_VALUE	' <4> Nr. of DATA-Bytes in a Block
  lSmBlockAdrIncs = FS_NUM_VARS_INIT_VALUE	' <4> Bloc-ADR Increment (i.d.R. identisch mit "RAM1_BYTES_IN_BLOCK")
  lSmBlockMask = FS_NUM_VARS_INIT_VALUE		' <2> Block-ADR Maske
  lSmBlockIMask = FS_NUM_VARS_INIT_VALUE	' <2> Block-ADR inverse Maske
  lSmBlockAdrSrl = FS_NUM_VARS_INIT_VALUE	' <1> Block-ADR Shifts-Right:  ADR --> to Block-Number (2**14 = 16384 = 16KByte)
  lSmNoOfBlocks = FS_NUM_VARS_INIT_VALUE	' <2> Total Number of "Blocks" in this SmartMedia Card
  lSmAdrHighBlock = FS_NUM_VARS_INIT_VALUE	' <4> Base ADR of highest Block
  lSmAdrEnd = FS_NUM_VARS_INIT_VALUE		' <4> End-ADR = 1.ADR hinter dem letzen Byte !!

  lSmNoOfInvBlocks = FS_NUM_VARS_INIT_VALUE		' No-of-Invalid Blocks (total)
  sSmInvBlks$ = FS_STR_VARS_INIT_VALUE

  sSmPageBuffer$ = FS_STR_VARS_INIT_VALUE
  sSmPgSpareBuf$ = FS_STR_VARS_INIT_VALUE
  sSmBlockBuffer$ = FS_STR_VARS_INIT_VALUE
  lSmPageSizeWithSpare = FS_NUM_VARS_INIT_VALUE

  sSmFindBuffer$ = FS_STR_VARS_INIT_VALUE

  lSmWritePrevHandle = SM_WRITE_BAD_HANDLE

  for wlIndx = 0 to SM_MAX_ZONE_NUM - 1
    laSmLogToPhysZonesSize(wlIndx) = FS_NUM_VARS_INIT_VALUE
  next

#ifdef SM_USE_FAST_LOG2PHYS_FUNCS
  sSmLogToPhys$ = FS_STR_VARS_INIT_VALUE
  sSmPhysToLog$ = FS_STR_VARS_INIT_VALUE
#else
  for wlIndx = 0 to SM_MAX_ZONE_NUM - 1
    for wlSecIndx = 0 to SM_MAX_LOG_BLOCK_NUM_IN_ZONE - 1
      laSmLogToPhys(wlIndx, wlSecIndx) = FS_NUM_VARS_INIT_VALUE
    next
  next
'' string sSmLogToPhysCopy$(16K)
#endif

  lSmLogToPhysNumOfZones = FS_NUM_VARS_INIT_VALUE
  lSmMaxZoneSizePhys = FS_NUM_VARS_INIT_VALUE
  lSmMaxZoneSizeLog = FS_NUM_VARS_INIT_VALUE

  lSmLastWriteError = FS_NUM_VARS_INIT_VALUE

  lSmCisBlockAddr = FS_NUM_VARS_INIT_VALUE
  sSmCisField$ = FS_STR_VARS_INIT_VALUE

  lDataAreaStart = FS_NUM_VARS_INIT_VALUE

#ifdef SM_WRITE_ON_BLOCK_SWITCH
  lSmInMemBlockStart = FS_NUM_VARS_INIT_VALUE
  lSmInMemBlockSize = FS_NUM_VARS_INIT_VALUE
#endif	'' SM_WRITE_ON_BLOCK_SWITCH
end
#endif	'' FS_SECURE_INIT


'' The "chk_fnam" function can now search only in the strings that are shorter than 256 bytes;
'' the SSFDC forum papers say that the minimal size of a SM page can not be less than 256 bytes,
'' this information is used here
sub lSMFindStringInMemBlock( string spSearchedString$; long lpStartAddr, &
  lpMemBlockSize, lpStepSize; var long lpvStringAddr )
  byte blIsAvailable, blIsFound
  long llCurAddr, llCurSize, llCurPagePtr, llSearchedStrLen, llPosInStr

  blIsFound = FALSE
  lpvStringAddr = -2
  llSearchedStrLen = len(spSearchedString$)
  if llSearchedStrLen > 255 or llSearchedStrLen > lpMemBlockSize then
    return
  endif
  llCurAddr = lpStartAddr
  llCurSize = lSmBytesInPage - mod( llCurAddr, lSmBytesInPage )
  if llCurSize > lpMemBlockSize then
    llCurSize = lpMemBlockSize
  endif
  while blIsFound = FALSE and llCurAddr - lpStartAddr < lpMemBlockSize
    call sSMReadString( llCurAddr, llCurSize, sSmFindBuffer$, blIsAvailable )
    if blIsAvailable = TRUE then
      llCurPagePtr = 0
      while blIsFound = FALSE and llCurPagePtr < llCurSize
''        llPosInStr = instr( spSearchedString$, sSmFindBuffer$, llCurPagePtr, 4 )
	llPosInStr = chk_fnam( sSmFindBuffer$, llCurPagePtr, lpStepSize, spSearchedString$, 0 )
        if llPosInStr >= 0 then
          lpvStringAddr = llCurAddr + llPosInStr
          blIsFound = TRUE
        endif
        llCurPagePtr = llCurPagePtr + 255 - llSearchedStrLen
      endwhile
    endif
    llCurAddr = llCurAddr + llCurSize
    if llCurAddr - lpStartAddr + lSmBytesInPage < lpMemBlockSize then
      llCurSize = lSmBytesInPage
    else
      llCurSize = lpMemBlockSize - ( llCurAddr - lpStartAddr )
    endif
  endwhile
end

#endif

